# functions for loading and manipulating phylogenetic trees
from Bio import Phylo

def bioPhyloToTupleTree(bpTree):
    '''Convert a biopython tree object to 4 tuple tree.'''
    return bioPhyloCladeToTupleTree(bpTree.root)

def bioPhyloCladeToTupleTree(clade):
    '''Convert a biopython clade object to 4 tuple tree.'''
    nm = clade.name
    br = clade.branch_length
    if clade.is_terminal():
        return (nm,(),(),br)
    else:
        lt = bioPhyloCladeToTupleTree(clade[0])
        rt = bioPhyloCladeToTupleTree(clade[1])
        return (nm,lt,rt,br)
    
def nodeCount(tree):
    '''How many nodes in tree?'''
    if tree[1]==():
        return 1
    else:
        return 1+ nodeCount(tree[1]) + nodeCount(tree[2])

def nodeList(tree):
    '''Return list of nodes in tree.'''
    if tree[1]==():
        return [tree[0]]
    else:
        return [tree[0]] + nodeList(tree[1]) + nodeList(tree[2])
    
def leafCount(tree):
    '''How many leaves in tree?'''
    if tree[1]==():
        return 1
    else:
        return leafCount(tree[1]) + leafCount(tree[2])

def leafList(tree):
    '''Return list of leaves in tree.'''
    if tree[1]==():
        return [tree[0]]
    else:
        return leafList(tree[1]) + leafList(tree[2])

def iNodeList(tree):
    '''Return list of internal nodes in tree.'''
    if tree[1]==():
        return []
    else:
        return [tree[0]] + iNodeList(tree[1]) + iNodeList(tree[2])
    
def subtree(tree,node):
    '''Return the subtree with node at its root. Assume node is in tree.'''
    if tree[0]==node:
        return tree
    elif tree[1]==():
        return None
    else:
        l=subtree(tree[1],node)
        r=subtree(tree[2],node)
        if l==None:
            return r
        else:
            return l
    
def isRootNode(tree,mrcaNum):
    '''Is mrcaNum the root node?'''
    # root node is tree[0] in our tuple trees
    return mrcaNum == tree[0]
    
def strTree2numTree(tree,counter):
    '''Given a tuple tree with nodes specified by strings, convert to
numbered nodes. Return tree with numbered nodes.'''
    if tree[1]==():
        return (counter,(),(),tree[3]),counter+1
    else:
        leftNumTree,counter=strTree2numTree(tree[1],counter)
        rightNumTree,counter=strTree2numTree(tree[2],counter)
        numTree=(counter,leftNumTree,rightNumTree,tree[3])
        return numTree,counter+1

def makeTreeD(tree1,tree2,treeD):
    '''Make a dictionary to convert from node names in tree1 to node names
in the identically shaped tree2.'''
    treeD[tree1[0]]=tree2[0]
    if tree1[1]==(): return
    else:
        makeTreeD(tree1[1],tree2[1],treeD)
        makeTreeD(tree1[2],tree2[2],treeD)
        return

def checkTree(bpTree):
    '''Check that a biopython tree is rooted and has named internal
nodes. Throw error if not. Returns None.'''

    # check its bifurcating. This actually ignors the root, but checks
    # for non-bifurcating nodes below that.
    if not bpTree.is_bifurcating():
        raise ValueError("This tree is not bifurcating.")
    
    # Now check to make sure there are only two clades at the root of
    # this tree. Otherwise, its going to mess up bioPhyloToTupleTree
    if len(bpTree.clade) > 2:
        raise ValueError("readTree requires a rooted tree with exactly two clades at the root. This input tree has more than two. This typically means that it is intended to be read as an unrooted tree.")

    # Make sure the internal nodes are labelled.
    predictedNumIntNodes = bpTree.count_terminals() - 1
    internalL = bpTree.get_nonterminals()
    if len(internalL) != predictedNumIntNodes or any((n.name==None for n in internalL)):
        raise ValueError("All the internal nodes of the input tree must have names.")
    
def readTree(filename):
    '''Read Newick tree from file and convert it to tuple format.'''

    bpTree = Phylo.read(filename, 'newick', rooted=True)

    checkTree(bpTree)
    
    stringTree=bioPhyloToTupleTree(bpTree)
    counter=0
    numTree,counter=strTree2numTree(stringTree,counter)

    # make dictionaries for converting between number and string strain names
    strainStr2NumD={}
    makeTreeD(stringTree,numTree,strainStr2NumD)
    strainNum2StrD={}
    makeTreeD(numTree,stringTree,strainNum2StrD)
    return numTree,strainStr2NumD,strainNum2StrD

def createSubtreeL(tree):
    '''Return a list containing all subtrees.'''
    if tree[1]==():
        return [tree]
    else:
        return [tree]+createSubtreeL(tree[1]) + createSubtreeL(tree[2])

def getNearestNeighborL(leaf,tree):
    '''Given a leaf and a tree, return a list containing the other leaf or
leaves which are most closely related to leaf.'''

    parent = getParent(leaf,tree) # assume leaf is in tree
    leafL = leafList(subtree(tree,parent))
    leafL.remove(leaf)
    return leafL
    
def getParent(leaf,tree):
    '''Return parent node of leaf.'''
    if tree[1] == ():
        # We've come too far, it wasn't there
        return None
    elif tree[1][0] == leaf or tree[2][0] == leaf:
        return tree[0]
    else:
        left = getParent(leaf,tree[1])
        right = getParent(leaf,tree[2])

        if left != None:
            return left
        else: return right
        
def tupleTree2Newick(tree):
    '''Convert a four tuple based tree (root,left,right,branchLen) into a
newick formated string.'''
    if tree[1]==():
        return str(tree[0])+":"+str(tree[3])
    else:
        leftStr=tupleTree2Newick(tree[1])
        rightStr=tupleTree2Newick(tree[2])
        return "("+leftStr+","+rightStr+"):"+str(tree[3])

def writeTree(tree,fileName):
    '''Write tree to fileName (in newick format).'''
    f=open(fileName,"w")
    f.write("("+tupleTree2Newick(tree)+");")
    f.close()



import sys,statistics,os,glob,random
from urllib import parse
from . import trees

def createIslandByStrainD(leafNodesL,strainNum2StrD,islandByNodeL,familiesO,geneNames,geneInfoD):
    '''Return a dict keyed by strain name. Values are lists of tuples
    (locIslandNum, locFamilyL) where locFamilyL is a list of tuples in
    the locus island present in that strain. locusFamily tuples are
    (locusFamily,[genes in locusFamily]).

    '''
    islandByStrainD = {}
    for leaf in leafNodesL:
        islandByStrainD[strainNum2StrD[leaf]]=[]

    # loop over every node in tree. examine each island. from each
    # extract the genes present in each strain and put in right entry
    # in islandByStrainD
    for mrcaNum in range(len(islandByNodeL)):
        for locIsland in islandByNodeL[mrcaNum]:

            # make dict to collect fams for each strain
            tempStrainD = {}
            for leaf in leafNodesL:
                tempStrainD[strainNum2StrD[leaf]]=[]

            # put LocusFamily number and genes tuple for each strain in tempStrainD
            for locFam in locIsland.iterLocusFamilies(familiesO):
                for leaf in leafNodesL:
                    geneNamesL=[geneNames.numToName(gene) for gene in locFam.iterGenesByStrain(leaf)]
                    if geneNamesL != []:
                        # only add if the family has some genes in this strain.
                        tempStrainD[strainNum2StrD[leaf]].append((locFam.locusFamNum,geneNamesL))


            # now make island tuple (minStart,island, locFamilyL) where
            # minStart is the lowest start coord we've seen in the
            # island, island is the island id, and locFamilyL is the thing
            # in tempStrainD
            for strain in islandByStrainD:
                # only add if the island is present in this strain

                if tempStrainD[strain] != []:
                    chrom,islandMedianMidpoint,islandMin,islandMax = getIslandPositions(tempStrainD[strain],geneInfoD,strainNum2StrD,locIsland.id,mrcaNum,strain)
                    locIslandT = (chrom,islandMedianMidpoint,islandMin,islandMax,mrcaNum,locIsland.id,tempStrainD[strain])
                    islandByStrainD[strain].append(locIslandT)
                    
    # sort each list in islandByStrainD by chrom and islandMedianMidpoint
    for strain in islandByStrainD:
        islandByStrainD[strain].sort(key=lambda x: x[:2])
    
    return islandByStrainD

def getIslandPositions(locFamilyL,geneInfoD,strainNum2StrD,locIslandID,mrcaNum,strain):
    '''Given a list of families (from a single island in a single strain),
return its chrom,start,end.
    '''
    chromL=[]
    islandMin=float('inf')
    islandMax=-float('inf')
    geneMidpointL=[]
    for locFam,geneL in locFamilyL:
        for gene in geneL:
            commonName,locusTag,descrip,chrom,start,end,strand=geneInfoD[gene]
            chromL.append(chrom)
            start = int(start)
            end = int(end)
            if start<islandMin:
                islandMin=start
            if end>islandMax:
                islandMax=end

            geneMidpointL.append(int((end-start)/2))
                
    # sanity check: all entries in chromL should be same
    if not all((c==chromL[0] for c in chromL)):
        print("Genes in island",locIslandID,"at mrca",strainNum2StrD[mrcaNum],"in strain",strain,"are not all on the same chromosome.",file=sys.stderr)

    islandMedianMidpoint = statistics.median(geneMidpointL)
    
    return chrom,islandMedianMidpoint,islandMin,islandMax

def orderedIslandsInStrain(strainName,islandByStrainD):
    '''returns a list of all the islands in the strain given, in the order they appear'''
    islandsInStrainL = islandByStrainD[strainName]
    chromStartIslandLL = []
    #for each island in the strain, order first by chromosome and then by island start pos
    for islandT in islandsInStrainL:
        chrom,_,start,_,_,locIslandNum,locFamilyL=islandT
        chromStartIslandLL.append([chrom,start,islandT])
    sortedChromStartIslandLL = sorted(chromStartIslandLL)
    #make an ordered list of just the islands without the other information
    sortedIslandsL = [x[2] for x in sortedChromStartIslandLL]
    return sortedIslandsL


def islandToBed(islandT,geneInfoD,tree,strainNum2StrD, islandColorD):
    '''Given a islandT (the values of islandByStrainD are lists of these)
convert into a string suitable for writing in a bed file. Return
this. Note that we're using the score field to color the genes in
IGB. So, we give genes in a island the same score, and give different
scores to adjacent islands. Counter keeps track of how many islands
we've done already.
    '''
    bedL=[]
    chrom,islandMedianMidpoint,islandMin,islandMax,mrcaNum,locIslandNum,locFamilyL = islandT
    locIslandID = 'locIsland_'+str(locIslandNum)
    score = str(islandColorD[str(locIslandNum)])
    
    # loop over families to get genes
    for locFam,geneL in locFamilyL:
        for gene in geneL:
            commonName,locusTag,descrip,chrom,start,end,strand=geneInfoD[gene]
            if commonName != '':
                Name=commonName
            else:
                Name=gene

            attributes = 'ID='+gene+';Name='+Name+';gene='+Name+';Note= | '+locIslandID+" | locFam_"+str(locFam)+" | mrca_"+strainNum2StrD[mrcaNum] + " | "+descrip
            
            bedL.append('\t'.join([chrom,start,end,Name,'0',str(strand),start,start,score,'1',str(int(end)-int(start)),'0',gene,attributes]))

    bedStr = '\n'.join(bedL)
    return bedStr
    
def writeStrainBed(islandByStrainD,geneInfoD,tree,strainNum2StrD,strain,bedFileName,islandColorD):
    '''For a given strain, Write bed file.'''
    f=open(bedFileName,'w')
    f.write('track name='+strain+' type=bedDetail visibility=full itemRgb="On" useScore=0 \n')
    orderedIslandsInStrainL = orderedIslandsInStrain(strain,islandByStrainD)
    for islandT in orderedIslandsInStrainL:
        bedStr = islandToBed(islandT,geneInfoD,tree,strainNum2StrD,islandColorD)
        f.write(bedStr+'\n')
    f.close()

def createAllBeds(islandByStrainD,geneInfoD,tree,strainNum2StrD,strainStr2NumD,paramD):

    bedFilePath = paramD['bedFilePath']
    potentialRgbL = paramD['potentialRgbL']
    bedNumTries = paramD['bedNumTries']

    # create a set of all leaf nodes outside the rootFocalClade
    allLeavesS = set(trees.leafList(tree))
    focalLeavesS = set(trees.leafList(trees.subtree(tree,strainStr2NumD[paramD['rootFocalClade']])))
    nodesOutsideFocalCladeS= allLeavesS - focalLeavesS
    
    # if directory for beds doesn't exist yet, make it
    bedDir = bedFilePath.split("*")[0]
    if glob.glob(bedDir)==[]:
        os.mkdir(bedDir)

    bedExtension = bedFilePath.split("*")[1]
    minIslandsMiscolored = 1000
    strainL = list(islandByStrainD.keys())
    for i in range(0,bedNumTries):
        random.shuffle(strainL)
        numberOfIslandsMiscolored,islandColorD  = createIslandColorD(strainL,nodesOutsideFocalCladeS,strainNum2StrD,potentialRgbL,islandByStrainD)
        if numberOfIslandsMiscolored<minIslandsMiscolored:
            bestColorD = islandColorD
            minIslandsMiscolored = numberOfIslandsMiscolored
    islandColorD = bestColorD
    numberOfIslandsMiscolored = minIslandsMiscolored
    
    for strain in islandByStrainD:
        bedFileName = bedDir+strain+bedExtension
        writeStrainBed(islandByStrainD,geneInfoD,tree,strainNum2StrD,strain,bedFileName,islandColorD)

    #print('Number of islands miscolored is '+str(numberOfIslandsMiscolored)+' after '+str(bedNumTries)+' tries.',file=sys.stderr)


def islandsNextToSameColorCount(islandByStrainD,islandColorD,nodesOutsideFocalCladeS,strainNum2StrD):
    '''counts the number of islands that are adjacent to the same color island'''
    miscolorCount = 0
    for strain in islandByStrainD:
        orderedIslandsInStrainL = orderedIslandsInStrain(strain,islandByStrainD)
        for islandIndex in range(1,len(orderedIslandsInStrainL)):
            mrcaNum = orderedIslandsInStrainL[islandIndex-1][-3]
            if strainNum2StrD[mrcaNum] not in nodesOutsideFocalCladeS:
                if islandColorD[str(orderedIslandsInStrainL[islandIndex][-2])] is islandColorD[str(orderedIslandsInStrainL[islandIndex-1][-2])]: miscolorCount+=1
                   
    return miscolorCount

def createIslandColorD(strainL,nodesOutsideFocalCladeS,strainNum2StrD,potentialRgbL,islandByStrainD):
    '''make the island color dictionary'''
    islandColorD = {}
    for strain in strainL:
        counter = 0
        orderedIslandsInStrainL = orderedIslandsInStrain(strain,islandByStrainD)
        for islandIndex in range(0,len(orderedIslandsInStrainL)):
            island = orderedIslandsInStrainL[islandIndex]
            #make a variable holding the previous island to compare colors
            if islandIndex != 0:
                prevIsland = orderedIslandsInStrainL[islandIndex-1]
                prevIslandNum = str(prevIsland[-2])
            mrcaNum = island[-3]
            islandNum= str(island[-2])

            # create score for coloring islands
            if strainNum2StrD[mrcaNum] in nodesOutsideFocalCladeS:
                score = '0,0,0'
                islandColorD[islandNum]=str(score)
            elif str(islandNum) in islandColorD:
                score = str(islandColorD[str(islandNum)])
            else:
                score = str(potentialRgbL[counter%len(potentialRgbL)])
                islandColorD[islandNum]=str(score)
                counter += 1

            #for all islands except the first one, see if the previous island is the same
            #color. if so, pick the next color and increment the counter
            if (islandIndex != 0) and (strainNum2StrD[mrcaNum] not in nodesOutsideFocalCladeS):
                if islandColorD[islandNum] is islandColorD[prevIslandNum]:
                    score = str(potentialRgbL[counter%len(potentialRgbL)])
                    islandColorD[islandNum]=str(score)
                    counter += 1

    numberOfIslandsMiscolored = islandsNextToSameColorCount(islandByStrainD,islandColorD,nodesOutsideFocalCladeS,strainNum2StrD)

    return numberOfIslandsMiscolored, islandColorD



# module for loading parameters.
import os

#### Base parameters

# These are parameters which users are not likely to need to change.
# We store them in a string of the same format as a parameters file.

baseParamStr = """

#### Blast ####

# blast command line (except for value for evalue as well as db,query
# and outfiles)
blastCLine = 'blastp -matrix BLOSUM62 -gapopen 11 -gapextend 1 -seg yes -outfmt 6 -evalue '

#### Scores calculation ####

# This helps determine the size of the arrays for our hash table of
# scores. It should be a value >= to 1. If one, it means the hash
# arrays have as many entries as there are score pairs.
hashArrayScaleFactor = 2

#### Family formation ####

# Family formation involves several thresholds. We determine these by
# looking at histograms of scores.
scoreHistNumBins = 80

## absMinRawThresholdForHomology
# A gene with a raw score below absMinRawThresholdForHomology will not
# be added to a family. If we cannot find a homologous peak in the
# histogram, we will just go with the following default.
defaultAbsMinRawThresholdForHomology = 0.5

# parameters for finding the homologous (right) peak in raw score
# histograms. These put restrictions on the peaks we can
# find. widthRelHeight specifies where we measure the width of a peak,
# as a proportion of the way down from the top.
homologRightPeakLimit = 1.0
widthRelHeight = 0.9

# case 1 (normal case)
homologPeakWidthCase1 = 0.05
homologRequiredProminenceCase1 = 0.4
homologLeftPeakLimitCase1 = 0.65

# case 2 (extreme prominence. But allow to be very narrow)
homologPeakWidthCase2 = 0
homologRequiredProminenceCase2 = 6
homologLeftPeakLimitCase2 = 0.90

# case 3 (wide width with low prominence)
homologPeakWidthCase3 = 0.25
homologRequiredProminenceCase3 = 0.10
homologLeftPeakLimitCase3 = 0.65

# parameters for finding the non-homologous (left) peak in raw score
# histograms.
nonHomologPeakWidth = 0.15
nonHomologPeakProminence = 1
nonHomologLeftPeakLimit = 0
nonHomologRightPeakLimit = 0.6

## synteny threshold parameters

# minSynThreshold is a synteny threshold below which we will not add a
# gene to a family. synAdjustThreshold is a synteny theshold above
# which we make a gene more likely to be added to a family (by bumping
# up it's raw score).

defaultMinSynThreshold = 0.6
defaultSynAdjustThreshold = 0.8

# These increments indicate how much to modify the left extreme of the
# synteny peak to get minSynThreshold or synAdjustThreshold
minSynThresholdIncrement = 0.15
synAdjustThresholdIncrement = 0.02

# finding the homologous (right) synteny peak
synLeftPeakLimit = 0.6
synRightPeakLimit = 1.0

# case 1 (High prominence, narrow peak. Close relatedness
synPeakWidthCase1 = 0.01
synRequiredProminenceCase1 = 2.0

# case 2 (wide width, low prominence. Distant relatedness.)
synPeakWidthCase2 = 0.10
synRequiredProminenceCase2 = 0.1


# We use syntenty scores to adjust similarity scores in family
# finding. In particular, if synteny is very good, we give the raw
# similarity score a little bump to increase the chances of the gene
# under consideration being added to the family. This parameter
# specifies the amount we multiply a rawScore by during this
# adjustment.
synAdjustExtent = 1.05

# When forming gene families within a single strain, we need a raw
# score threshold to determine which genes are similar enough to get
# in the same family. To get this threshold, we identify the nearest
# neighbors of the strain in question, and get the average score in
# core genes across these. We then multiply by this adjustment
# parameter to get the value we use as a threshold.
singleStrainFamilyThresholdAdjust = 0.5

#### LocusIsland formation ####

# geneProximityRange tells us how far out we want to go from each gene
# in recording proximity for geneProximityD
geneProximityRange = 2

# In deciding whether to merge two islands, we judge partly based on
# the proximity of their genes. proximityThreshold defines what
# proximity is. A proximityThreshold of 1 means adjacent genes
proximityThresholdMerge1 = 1

# rscThreshold of 0 means that we merge islands if their rscore is 0
# or above
rscThresholdMerge1 = 0

# maxClusterSize is the maximum size that we make clusters of islands
# in the first step of the merging process
maxClusterSize = 50


#### Visualization and analysis ####


bedNumTries = 100 # number of random tries to find best coloring for islands

# The possible rgb values for bed files is based on the list below. 
potentialRgbL = ['245,130,48', '188,143,14','0,102,0','230,26,135','0,0,128', '145,30,180','0,255,255','128,0,0','0,255,0', '255,0,255','240,230,140','32,178,170','240,128,128','255,0,0','100,149,237','60,179,113','0,255,130','128,128,128','94,94,94','102,51,0']

"""


#### Functions
def createParametersD(baseParamStr,paramFN):
    '''Create and return a parameters dictionary. First parses the string
passed in as baseParamStr. This consists of parameters users are less
likely to modify. Then adds user specific parameters contained in the
file paramFN. Note that because the user parameters are put in the
parameters dictionary second, it is possible for a user to override
one of the base parameters simply by including that parameter in their
parameter file.

    '''
    paramD={}
    baseParamL = baseParamStr.split('\n')
    paramD = addParametersToD(baseParamL,paramD)

    f=open(paramFN,'r')
    userParamL = f.read().split('\n')
    paramD = addParametersToD(userParamL,paramD)
    f.close()
    
    return paramD
    
def addParametersToD(paramL,paramD):
    '''Given a list of lines (e.g. from a parameters file) add parameters
to paramD. Each line is a string. Some will be comments or blank
lines, and others will be python assignment statements. We use the
assignment statements to create entries in paramD.
    '''

    for s in paramL:
        if s == '' or s.lstrip() == '' or s.lstrip()[0]=='#':
            continue
        # it's not a blank line, a line with only whitespace or a
        # comment.  so it must be an assignment statement.
        key,value = s.rstrip().split('=')
        key = key.strip()
        value = value.strip()
        paramD[key] = eval(value)
    
    return paramD    
    
def loadFileNameMapD(fileNameMapFN,genbankFileList=None):
    '''Create a dictionary with mappings between genbank file names and
the human readable names we use in the tree. If fileNameMapFN contains
a string, we load that file and construct the mappings based on this.
Expects file with one species per line: genbank name + white space +
human name. If fileNameMapFN is None, we create the mappings between
the full file names (with path and extension) and the stem of the file
name, which in this case should correspond to what is in the input
tree.
    '''
    fileNameMapD = {}
    if fileNameMapFN == None:
        for fullPathFN in genbankFileList:
            fn = os.path.split(fullPathFN)[-1]
            stem = os.path.splitext(fn)[0]
            fileNameMapD[fn] = stem
    else:
        f = open(fileNameMapFN,'r')
        while True:
            s = f.readline()
            if s == '':
                break
            elif s[0].isspace():
                continue
            genbankStem,human = s.rstrip().split()
            fileNameMapD[genbankStem] = human
    return fileNameMapD



"""Provides the entry point to xenoGI's functionality."""
__version__ = "2.1.0"
import sys, glob, os
from . import parameters,genbank,blast,trees,genomes,Score,scores,Family,families,islands,analysis,islandBed

def main():
    
    #### check command line
    try:
        assert(len(sys.argv) == 3)
        paramFN=sys.argv[1]
        task = sys.argv[2]
        assert(task in ['parseGenbank', 'runBlast', 'calcScores', 'makeFamilies', 'makeIslands', 'printAnalysis', 'createIslandBed', 'plotScoreHists', 'interactiveAnalysis', 'runAll', 'version', 'debug', 'simValidation'])
    
    except:
        print(
            """
   Exactly two arguments required.
      1. A path to a parameter file.
      2. The task to be run which must be one of: parseGenbank, runBlast, calcScores, makeFamilies, makeIslands, printAnalysis, createIslandBed, plotScoreHists, interactiveAnalysis, runAll or version.

   For example: 
      xenoGI params.py parseGenbank
"""
            ,file=sys.stderr)
        sys.exit(1)

        
    #### load parameters and some other data we'll use below
    paramD = parameters.createParametersD(parameters.baseParamStr,paramFN)
        
    #### parseGenbank
    if task == 'parseGenbank':
        parseGenbankWrapper(paramD)
        
    #### runBlast
    elif task == 'runBlast':
        runBlastWrapper(paramD)
        
    #### calcScores
    elif task == 'calcScores':
        calcScoresWrapper(paramD)

    #### makeFamilies
    elif task == 'makeFamilies':
        makeFamiliesWrapper(paramD)
        
    #### makeIslands
    elif task == 'makeIslands':
        makeIslandsWrapper(paramD)

    #### printAnalysis
    elif task == 'printAnalysis':
        printAnalysisWrapper(paramD)

    #### createIslandBed
    elif task == 'createIslandBed':
        createIslandBedWrapper(paramD)

    #### interactiveAnalysis
    elif task == 'interactiveAnalysis':
        interactiveAnalysisWrapper(paramD)
        
    #### plotScoreHists
    elif task == 'plotScoreHists':
        plotScoreHistsWrapper(paramD)

    #### runAll
    elif task == 'runAll':
        parseGenbankWrapper(paramD)
        runBlastWrapper(paramD)
        calcScoresWrapper(paramD)
        makeFamiliesWrapper(paramD)
        makeIslandsWrapper(paramD)
        printAnalysisWrapper(paramD)
        createIslandBedWrapper(paramD)

    #### version
    elif task == 'version':
        print("xenoGI",__version__)
        
    #### debug
    elif task == 'debug':
        debugWrapper(paramD)

    #### validation via simulation
    elif task == 'simValidation':
        simValidationWrapper(paramD)
        
######## Task related functions

def parseGenbankWrapper(paramD):
    """Wrapper running stuff to parse genbank files."""
    genbankFileList=glob.glob(paramD['genbankFilePath'])

    # if directory for fastas doesn't exist yet, make it
    fastaDir = paramD['fastaFilePath'].split('*')[0]
    if glob.glob(fastaDir)==[]:
        os.mkdir(fastaDir)

    fileNameMapD = parameters.loadFileNameMapD(paramD['fileNameMapFN'],genbankFileList)

    # parse
    genbank.parseGenbank(paramD,fastaDir,genbankFileList,fileNameMapD)

def runBlastWrapper(paramD):
    """Wrapper to blast all genome files against each other."""
    dbFileL=blast.getDbFileL(paramD['fastaFilePath'],paramD['treeFN'])
    blast.runBlast(dbFileL,dbFileL,paramD)

    
def loadMiscDataStructures(paramD):
    """Creates a few data structures that are used in multiple tasks."""

    tree,strainStr2NumD,strainNum2StrD = trees.readTree(paramD['treeFN'])
    
    # an object for gene name conversions
    geneNames = genomes.geneNames(paramD['geneOrderFN'],strainStr2NumD,strainNum2StrD)

    subtreeL=trees.createSubtreeL(tree)
    subtreeL.sort()
    geneOrderT=genomes.createGeneOrderTs(paramD['geneOrderFN'],geneNames,subtreeL,strainStr2NumD)

    return tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT

def calcScoresWrapper(paramD):
    """Wrapper running stuff to calculate scores."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)

    # object for storing scores
    scoresO=Score.Score()
    scoresO.initializeDataAttributes(paramD['blastFilePath'],geneNames,strainStr2NumD)

    ## similarity scores
    scoresO = scores.calcRawScores(paramD,geneNames,scoresO)

    ## synteny scores
    scoresO = scores.calcSynScores(scoresO,geneNames,geneOrderT,paramD,tree)

    ## core synteny scores
    strainNamesL=sorted([strainNum2StrD[leaf] for leaf in trees.leafList(tree)])
    scoresO = scores.calcCoreSynScores(scoresO,strainNamesL,paramD,geneNames,geneOrderT)
    
    # write scores to file
    scores.writeScores(scoresO,geneNames,paramD['scoresFN'])
    
def makeFamiliesWrapper(paramD):
    """Wrapper to create gene families."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    ## read scores
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)
    aabrhL = scores.loadOrthos(paramD['aabrhFN'])

    ## make gene families
    familyFormationSummaryF = open(paramD['familyFormationSummaryFN'],'w')
    familiesO = families.createFamiliesO(tree,strainNum2StrD,scoresO,geneNames,aabrhL,paramD,subtreeL,familyFormationSummaryF)
    familyFormationSummaryF.close()

def makeIslandsWrapper(paramD):
    """Wrapper to create islands"""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    
    ## read scores
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)

    ## read gene families
    familiesO = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)

    ## group gene families into islands
    islandFormationSummaryF = open(paramD['islandFormationSummaryFN'],'w')
    locusIslandByNodeLMerged = islands.makeLocusIslands(geneOrderT,geneNames,subtreeL,tree,paramD,familiesO,strainStr2NumD,strainNum2StrD,islandFormationSummaryF)
    islandFormationSummaryF.close()

def printAnalysisWrapper(paramD):
    """Wrapper to run analysis."""

    ## setup output directory and file names
    # if directory for analysis doesn't exist yet, make it
    analDir = paramD['analysisFilePath'].split("*")[0]
    if glob.glob(analDir)==[]:
        os.mkdir(analDir)

    islandSummaryStem = paramD['islandsSummaryStem']
    analExtension = paramD['analysisFilePath'].split("*")[1]
    islandsSummaryFN = os.path.join(analDir,islandSummaryStem+analExtension)

    genesFNstem = os.path.join(analDir,paramD['genesFNstem'])
    
    ## load stuff
    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    islandByNodeL=islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)
    nodesL=trees.nodeList(tree)
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])
    familiesO = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)

    ## analysis

    # Print out all islands
    islandsOutF = open(islandsSummaryFN,'w')
    analysis.vPrintAllLocusIslands(islandByNodeL,tree,paramD['rootFocalClade'],subtreeL,familiesO,strainStr2NumD,strainNum2StrD,geneNames,geneInfoD,islandsOutF)
    islandsOutF.close()

    # Print species files with all the genes, grouped by contig
    gene2FamIslandD = analysis.createGene2FamIslandD(islandByNodeL,familiesO)
    analysis.printSpeciesContigs(geneOrderT,genesFNstem,analExtension,geneNames,gene2FamIslandD,geneInfoD,familiesO,strainNum2StrD)

def createIslandBedWrapper(paramD):
    """Wrapper to make output bed files."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    
    leafNodesL = trees.leafList(tree)
    familiesO = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)
    islandByNodeL = islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])    

    # get islands organized by strain
    islandByStrainD = islandBed.createIslandByStrainD(leafNodesL,strainNum2StrD,islandByNodeL,familiesO,geneNames,geneInfoD)

    islandBed.createAllBeds(islandByStrainD,geneInfoD,tree,strainNum2StrD,strainStr2NumD,paramD)

def plotScoreHistsWrapper(paramD):
    """Wrapper to make pdf of histograms of scores."""

    import matplotlib.pyplot as pyplot
    from matplotlib.backends.backend_pdf import PdfPages
    
    numBins = 80 # num bins in histograms
    tree,strainStr2NumD,strainNum2StrD = trees.readTree(paramD['treeFN'])
    geneNames = genomes.geneNames(paramD['geneOrderFN'],strainStr2NumD,strainNum2StrD)
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)
    
    def scoreHists(scoresFN,outFN,strainNum2StrD,numBins,geneNames,scoreType):
        '''Read through a scores file, and separate into all pairwise comparisons. Then plot hist of each.'''

        # currently, this seems to require a display for interactive
        # plots. would be nice to make it run without that...

        scoresO = scores.readScores(scoresFN,geneNames)

        pyplot.ioff() # turn off interactive mode
        with PdfPages(outFN) as pdf:
            for strainPair in scoresO.getStrainPairs():
                fig = pyplot.figure()
                scoresL = list(scoresO.iterateScoreByStrainPair(strainPair,scoreType))
                pyplot.hist(scoresL,bins=numBins, density = True)
                pyplot.title(strainNum2StrD[strainPair[0]]+'-'+strainNum2StrD[strainPair[1]])
                pdf.savefig()
                pyplot.close()

    # plot histograms
    for scoreType,outFN in [('rawSc','rawSc.pdf'),('synSc','synSc.pdf'),('coreSynSc','coreSynSc.pdf'),]:
    
        scoreHists(paramD['scoresFN'],outFN,strainNum2StrD,numBins,geneNames,scoreType)

    
def interactiveAnalysisWrapper(paramD):
    """Enter interactive mode."""

    ## Set up the modules a bit differently for interactive mode
    import code,sys
    from xenoGI.analysis import createGene2FamIslandD,printScoreMatrix,matchFamilyIsland,printIslandNeighb,vPrintLocusIslandsAtNode,coreNonCoreCtAtNode,printOutsideFamilyScores

    ## Wrapper analysis functions. For convenience these assume a
    ## bunch of global variables.
    
    def printFam(familyNum,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print relevant info on
    a family. For ease of use, we take only one argument, assuming all the
    other required stuff is available at the top level. familyNum is the
    numerical identifier of a family.
        '''
                    
        print("Family",familyNum,file=fileF)
        # print out the locus families
        for lfO in familiesO.getFamily(familyNum).getLocusFamilies():
            print("    LocusFamily",lfO.getStr(strainNum2StrD,geneNames," "),file=fileF)
        
        # print("Family error score (count of possibly misassigned genes):",familiesO[familyNum].possibleErrorCt,file=fileF)

        print(file=fileF)
        print("Matrix of raw similarity scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familiesO,geneNames,scoresO,'rawSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of core synteny scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familiesO,geneNames,scoresO,'coreSynSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of synteny scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familiesO,geneNames,scoresO,'synSc',fileF)
        print(file=fileF)
        print(file=fileF)

        printOutsideFamilyScores(familyNum,subtreeL,familiesO,geneNames,scoresO,fileF)
        print(file=fileF)
        print(file=fileF)

    def findIsland(searchStr,fileF=sys.stdout):
        '''Print the gene, LocusIsland family and LocusFamily associated with
    searchStr. This is a wrapper that assumes various required objects
    are present at the top level.
        '''
        L=matchFamilyIsland(geneInfoD,geneNames,gene2FamIslandD,searchStr)
        for geneName,locusIslandNum, famNum, locusFamNum in L:
            print("<gene:"+str(geneName)+">","<locIsl:"+str(locusIslandNum)+">","<fam:"+str(famNum)+">","<locFam:"+str(locusFamNum)+">",file=fileF)


    def printIsland(locusIslandNum,synWSize,fileF=sys.stdout):
        '''Print a LocusIsland and its genomic context in each species. We
        include synWSize/2 genes in either direction beyond the locus island.
        '''
        printIslandNeighb(locusIslandNum,synWSize,subtreeL,islandByNodeL,familiesO,geneOrderT,gene2FamIslandD,geneInfoD,geneNames,strainNum2StrD,fileF)


    def printIslandsAtNode(nodeStr,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print all the LocusIslands
    at a particular node in the tree. For ease of use, we take only a node
    number as argument, assuming all the other required stuff is available
    at the top level.
        '''
        node = strainStr2NumD[nodeStr]
        vPrintLocusIslandsAtNode(islandByNodeL[node],subtreeL,familiesO,strainNum2StrD,geneNames,geneInfoD,fileF)


    def printCoreNonCoreByNode(fileF=sys.stdout):
        '''For each node in the focal clade, print the number of core and
    non-core families. That is, for that node, we get all families present
    in descendent species. We then look at their mrcas. Those with an mrca
    below the node in question are non-core, others are core.'''

        focTree = trees.subtree(tree,strainStr2NumD[paramD['rootFocalClade']])
        focInodesL=trees.iNodeList(focTree)
        familyByNodeL=createFamilyByNodeL(geneOrderT,gene2FamD)

        rowL=[]
        rowL.append(['Node','Core','Non-Core','Total','% Non-Core'])
        rowL.append(['----','----','--------','-----','----------'])
        for node in focInodesL:
            nonCore,core=coreNonCoreCtAtNode(tree,node,familyByNodeL,familiesO)
            rowL.append([strainNum2StrD[node],str(core),str(nonCore),str(core+nonCore),str(format(nonCore/(core+nonCore),".3f"))])
        printTable(rowL,fileF=fileF)
        print(file=fileF)
        return


    ## Load data
    
    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    nodesL=trees.nodeList(tree)
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])
    familiesO = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)
    islandByNodeL=islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)
    gene2FamIslandD = createGene2FamIslandD(islandByNodeL,familiesO)
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)
    scoresO.createNodeConnectL(geneNames) # make nodeConnectL attribute
    
    code.interact(local=locals())

def debugWrapper(paramD):
    '''Take us into interactive mode for debugging.'''

    ## Set up the modules a bit differently for interactive mode
    import code,sys
    from .xenoGI import parameters,trees,genomes,families,islands,analysis,Score,scores

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
        
    code.interact(local=locals())

def simValidationWrapper(paramD):
    '''Take us into interactive mode for running validation with
simulations. This is not really intended for the end user...'''

    ## Set up the modules a bit differently for interactive mode
    import sys
    from .xenoGI import parameters,trees,genomes,families,islands,analysis
    sys.path.insert(0,'/data/bushlab/htrans/dupDelProj/simCode')
    import validationSim

    moduleT=(parameters,trees,genomes,families,islands,analysis)
    
    xgiIslandGeneSetByNodeL,simIslandGeneSetByNodeL=validationSim.runValidation(moduleT,'simParams.py','params.py')



def load(filename):
    """Load fasta or multifasta, return list of tuples (header,seq)."""
    f=open(filename,"r")
    header=None
    outL=[]
    tempSeqL=[]
    while True:
        Str=f.readline()
        if Str=="":
            outSeq="".join(tempSeqL)
            tempSeqL=[]
            outSeq="".join(outSeq.split()) # remove all whitespace
            outL.append((header,outSeq))
            break
        if Str[0]==">":
            # this is new header, put together previous header,seq, and move one
            if header!=None:
                outSeq="".join(tempSeqL)
                tempSeqL=[]
                outSeq="".join(outSeq.split()) # remove all whitespace
                outL.append((header,outSeq))
            header=Str[:-1]
        else:
            # this is a seq line,
            tempSeqL.append(Str)
    f.close()
    return(outL)    



import sys
from multiprocessing import Pool
from . import trees
from . import genomes
from . import analysis
from .Island import *
from .Family import *
import math 
    
## Main function  
    
def makeLocusIslands(geneOrderT,geneNames,subtreeL,tree,paramD,familiesO,strainStr2NumD,strainNum2StrD, outputSummaryF):
    '''Parallelized wrapper to merge locus islands at different nodes.'''

    numThreads = paramD['numThreads']
    rootFocalClade = paramD['rootFocalClade']
    islandOutFN = paramD['islandOutFN']
    geneProximityRange = paramD['geneProximityRange']
    proximityThresholdMerge1 = paramD['proximityThresholdMerge1']
    rscThresholdMerge1 = paramD['rscThresholdMerge1']
    maxClusterSize = paramD['maxClusterSize']
    
    geneProximityD = genomes.createGeneProximityD(geneOrderT,geneProximityRange )
    locIslByNodeL=createLocIslByNodeL(familiesO,tree)
    numIslandsAtEachNodeAtStartL = [len(L) for L in locIslByNodeL]
    focalNodesL = getFocalNodesInOrderOfNumDescendants(tree,strainStr2NumD,rootFocalClade)

    ##  Merge in clusters

    locusIslandClusterL,singletonClusterL = createLocusIslandClusters(locIslByNodeL,focalNodesL,subtreeL,familiesO,geneProximityD,geneProximityRange,maxClusterSize)

    # create argumentL to be passed to p.map and mergeLocIslandsAtNode
    argumentL = []
    for clusterL in locusIslandClusterL:
        argumentL.append((clusterL,geneProximityD,proximityThresholdMerge1,rscThresholdMerge1,subtreeL[clusterL[0].mrca],familiesO))
    p=Pool(numThreads)
    mergedL = p.map(mergeLocIslandsAtNode, argumentL) # run it

    # update locIslByNodeL with the merged nodes
    locIslByNodeL = updateIslandByNodeLEntries(locIslByNodeL,focalNodesL,mergedL)

    ## Family improvement

    # not currently doing this
    
    ##  Merge at mrca nodes
    argumentL = []
    for mrcaNode in focalNodesL:
        argumentL.append((locIslByNodeL[mrcaNode],geneProximityD,proximityThresholdMerge1,rscThresholdMerge1,subtreeL[mrcaNode],familiesO))
    p=Pool(numThreads)
    mergedL = p.map(mergeLocIslandsAtNode, argumentL) # run it

    # add the islands that were identified as singleton clusters
    mergedL.extend(singletonClusterL)
    
    # again update locIslByNodeL
    locIslByNodeL = updateIslandByNodeLEntries(locIslByNodeL,focalNodesL,mergedL)

    numIslandsAtEachNodeAtEndL = [len(L) for L in locIslByNodeL]

    ## Summary and output

    # print summary of merging
    printSummary(focalNodesL,strainNum2StrD,numIslandsAtEachNodeAtStartL,numIslandsAtEachNodeAtEndL,outputSummaryF)
    
    # write islands
    writeIslands(locIslByNodeL,strainNum2StrD,islandOutFN)

    return locIslByNodeL

## Support functions

def createLocIslByNodeL(familiesO,tree):
    '''Create locus islands, one family each. Initially store islands
separately by mrca in a list of lists. (The index of the outer list
corresponds to the mrca)
    '''
    locIslByNodeL=[[] for i in range(trees.nodeCount(tree))]

    for lfO in familiesO.iterLocusFamilies():
        liO = LocusIsland(lfO.locusFamNum, lfO.lfMrca, [lfO.locusFamNum])
        locIslByNodeL[liO.mrca].append(liO)

    # sort each list by island number
    for i in range(len(locIslByNodeL)):
        locIslByNodeL[i].sort(key=lambda x: x.id)
    
    return locIslByNodeL

def getFocalNodesInOrderOfNumDescendants(tree,strainStr2NumD,rootFocalClade):
    '''Get a list of all the nodes in the rootFocalClade. Then sort these
according to the number of leaves descending from them, biggest
first.'''

    focalSubtree = trees.subtree(tree,strainStr2NumD[rootFocalClade])
    focalNodesL = []
    for node in trees.nodeList(focalSubtree):
        # go through every node in the focal clade
        numDescend = trees.leafCount(trees.subtree(tree,node)) # count descendants
        focalNodesL.append((node,numDescend))
    focalNodesL.sort(key=lambda x: x[1],reverse=True) # sort on 2nd pos, biggest first
    # extract and return only the nodes
    return [node for node,numDescend in focalNodesL]

## Cluster formation

def createLocusIslandClusters(locIslByNodeL,focalNodesL,subtreeL,familiesO,geneProximityD,proximityThreshold,maxClusterSize):
    '''For every node in the focal clade, take the set of single family
LocusIslands in locIslByNodeL. Break this up into smaller
clusters based on the chromosomal distances between members of the
familes. Each resulting cluster should contain LocusIslands that we
are likely to merge.'''

    locusIslandClustersL = [] #holds non-singleton clusters 
    singletonClustersL = [] #holds all of the islands from all MRCA nodes that are returned in clusters only containing one island 
    
    #loop through all of the mrcaNode lists 
    for mrcaNode in focalNodesL:

        subtree = subtreeL[mrcaNode]
        islandsAtMrcaNodeL = locIslByNodeL[mrcaNode] 
        
        mrcaClustersL,mrcaSingletonClustersL = createMrcaNodeClusters(islandsAtMrcaNodeL,familiesO,subtree,geneProximityD,proximityThreshold,maxClusterSize)

        singletonClustersL.extend(mrcaSingletonClustersL)
        
        locusIslandClustersL.extend(mrcaClustersL)
    
    return locusIslandClustersL,singletonClustersL

def createMrcaNodeClusters(islandsAtMrcaNodeL,familiesO,subtree,geneProximityD,proximityThreshold,maxClusterSize):
    
    '''Takes in a Mrca node and other node information 
    and clusters the nodes within the mrca node that are likely 
    to merge'''

    mrcaClustersL = []
    mrcaSingletonClustersL = []
    
    while islandsAtMrcaNodeL:

        # clusterL is the ultimate destination for cluster
        # members. onDeckL is where we put islands that will be in the
        # cluster, but which haven't been used as a seed to search for
        # other cluster members yet.
        clusterL = [] 
        onDeckL = [islandsAtMrcaNodeL.pop()]
        
        clusterL,islandsAtMrcaNodeL = populateCluster(clusterL,onDeckL,islandsAtMrcaNodeL,familiesO,subtree,geneProximityD,proximityThreshold,maxClusterSize)
        
        if len(clusterL) > 1:
            mrcaClustersL.append(clusterL)
        else:
            mrcaSingletonClustersL.append(clusterL)

    return mrcaClustersL,mrcaSingletonClustersL


def populateCluster(clusterL,onDeckL,islandsAtMrcaNodeL,familiesO,subtree,geneProximityD,proximityThreshold,maxClusterSize):
    '''Populate clusterL by searching for islands in
islandsAtMrcaNodeL. onDeckL consists of islands that will be added to
clusterL, but which need to be used to search first.

    '''

    while onDeckL:
        
        searchSeedLocIslO = onDeckL.pop()
        # this LocIsl should only have one LocFam
        searchSeedLocFamO = list(searchSeedLocIslO.iterLocusFamilies(familiesO))[0]

        tempFoundL=[]
        for liO in islandsAtMrcaNodeL:
            locFamO = list(liO.iterLocusFamilies(familiesO))[0]

            proxB = proximitySubtree(searchSeedLocFamO,locFamO,geneProximityD,proximityThreshold,subtree)
            if proxB:
                tempFoundL.append(liO)

        # done searching, add to final list
        clusterL.append(searchSeedLocIslO)

        # add the things we found, 1 by 1
        for liO in tempFoundL:
            onDeckL.append(liO)
            islandsAtMrcaNodeL.remove(liO)
            
            if len(clusterL) + len(onDeckL) == maxClusterSize:
                # no more room. Stop now, leaving the rest of
                # tempFoundL in islandsAtMrcaNodeL
                clusterL.extend(onDeckL)
                return clusterL,islandsAtMrcaNodeL

    return clusterL,islandsAtMrcaNodeL
        
def proximitySubtree(lfam0,lfam1,geneProximityD,proximityThreshold,subtree):
    '''Given two gene families with the same mrcaNode, return boolean
indicating whether any of their genes are within proximityThreshold of
each other.
    '''
    if subtree[1] == ():
        return proximity(lfam0,lfam1,geneProximityD,proximityThreshold,subtree[0])
    else:
        left = proximitySubtree(lfam0,lfam1,geneProximityD,proximityThreshold,subtree[1])
        right = proximitySubtree(lfam0,lfam1,geneProximityD,proximityThreshold,subtree[2])
        return left or right

def updateIslandByNodeLEntries(locIslByNodeL,focalNodesL,mergedL):
    '''Given a list mergedL containing merged clusters of LocusIslands,
put back in locIslByNodeL. mergedL clusters all have mrca within
focalNodesL. We first blank the corresponding mrca location in
locIslByNodeL, then add the contents of mergedL.'''
    
    # blank out the entries in locIslByNodeL which are within the root focal clade
    for mrcaNode in focalNodesL:
        locIslByNodeL[mrcaNode] = []

    # fill them with LocusIslands from mergedL
    for clusterL in mergedL:
        if clusterL != []:
            locIslByNodeL[clusterL[0].mrca].extend(clusterL)

    return locIslByNodeL
    
## Iterative merging

def mergeLocIslandsAtNode(argT):
    '''Given a list of locus islands at one node (locusIslandL)
iteratively merge until there are no more pairwise scores above
threshold. rscThreshold represents the threshold below which we no
longer merge islands.
    '''

    locusIslandL,geneProximityD,proximityThreshold,rscThreshold,subtree,familiesO = argT
    
    if len(locusIslandL) < 2:
        # nothing to merge
        return locusIslandL

    # Pre-calculate costDiff scores between all families in locusIslandL
    lFamNumL = []
    for liO in locusIslandL:
        # we only need to consider first and last loc fams
        lFamNumL.append(liO.locusFamilyL[0]) # add first
        if len(liO) > 1:
            # this li has more than one loc fam, add last as well
            lFamNumL.append(liO.locusFamilyL[-1])
        
    costDiffD = costDiffDict((lFamNumL,familiesO,geneProximityD,proximityThreshold,subtree))
    
    # create initial scoreD
    scoreD = createScoreD(locusIslandL,costDiffD)

    # Merge
    while True:
        sc,locIslandPairT,orientation=maxScore(scoreD)
        if sc < rscThreshold:
            break

        ind0,li0 = searchLocIslandsByID(locusIslandL,locIslandPairT[0])
        ind1,li1 = searchLocIslandsByID(locusIslandL,locIslandPairT[1])

        li0.merge(li1,orientation)

        # delete li1
        del locusIslandL[ind1]

        # remove all scores in scoreD that involve li0 or li1
        delScores(scoreD,li0.id,li1.id)

        # calculate new scores for li0 against all other islands
        addScores(scoreD,li0,locusIslandL,costDiffD)
        
    return locusIslandL

def maxScore(scoreD):
    '''Find the entry with the highest score in scoreD, and return the
corresponding key.'''
    bestSc=-float('inf')
    bestLocIslandPairT=()
    bestOrientation = 0
    for locIslandPairT,(sc,orientation) in scoreD.items():
        if sc > bestSc:
            bestSc=sc
            bestLocIslandPairT = locIslandPairT
            bestOrientation = orientation
    return bestSc,bestLocIslandPairT,bestOrientation

def delScores(scoreD,li0ID,li1ID):
    '''Given two locus island ids from newly merged nodes, delete any entries in
score D that come from them.'''
    # find ones to delete
    toDelL=[]
    for key in scoreD.keys():
        if li0ID in key or li1ID in key:
            toDelL.append(key)

    # now delete them
    for key in toDelL:
        del scoreD[key]

def addScores(scoreD,li0,locusIslandL,costDiffD):
    '''Get scores for locus island li0 against all other locus islands and add to
scoreD.'''
    for locIsl in locusIslandL:
        if locIsl.id != li0.id:
            storeScore(li0,locIsl,costDiffD,scoreD)

def searchLocIslandsByID(listOfLocIslands,id):
    '''Search for a locus island with id equal to id. Return the index of the
first we find, and None if their isn't one.
    '''
    for i,isl in enumerate(listOfLocIslands):
        if isl.id == id: return i,isl
    return None,None

## Scoring

def costDiffDict(argT):
    '''Create a dictionary of costDiff scores between all pairs of locus
families in lFamNumL.'''

    lFamNumL,familiesO,geneProximityD,proximityThreshold,subtree=argT

    cdD={}
    for i in range(len(lFamNumL)-1):
        lf1 = familiesO.getLocusFamily(lFamNumL[i])
        for j in range(i+1,len(lFamNumL)):
            lf2 = familiesO.getLocusFamily(lFamNumL[j])
            cdsc = costDiff(lf1,lf2,geneProximityD,proximityThreshold,subtree)
            cdD[(lf1.locusFamNum,lf2.locusFamNum)] = cdsc
            cdD[(lf2.locusFamNum,lf1.locusFamNum)] = cdsc
    return cdD

def costDiff(lfam1,lfam2,geneProximityD,proximityThreshold,subtree):
    '''Given two families calculate the difference in rcost depending on
whether we assume the root is not proximate or proximate. lfam1 and
lfam2 are family tuples specifying the genes present in a family.
    '''
    memoD = {}
    t=rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree,True,memoD)
    f=rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree,False,memoD)
    return(f-t)

def createScoreD(locusIslandL,costDiffD):
    '''Create dictionary of scores between all locus islands at a single node
(and thus with the same mrca).'''
    scoreD={}
    for i in range(len(locusIslandL)-1):
        for j in range(i+1,len(locusIslandL)):
            storeScore(locusIslandL[i],locusIslandL[j],costDiffD,scoreD)
    return scoreD

def storeScore(li0,li1,costDiffD,scoreD):
    '''Calculate and store score in scoreD. We follow convention that key
in scoreD should always have the lower island id first.'''
    if li0.id < li1.id:
        key = li0.id,li1.id
        score,orientation=rscore(li0,li1,costDiffD)
        # rscore returns different things depending on order
        # so we must be consistent with what we do in key.
    else:
        key = li1.id,li0.id
        score,orientation=rscore(li1,li0,costDiffD)
        
    scoreD[key]=(score,orientation)
    
def proximity(lfam1,lfam2,geneProximityD,proximityThreshold,strain):
    '''Return True if any of the genes at strain from lfam1 are within
proximityThreshold of genes at that strain for lfam2. proximityThreshold
is measured in genes, e.g. 1 means the adjacent gene.
    '''
    for gn1 in lfam1.iterGenesByStrain(strain):
        for gn2 in lfam2.iterGenesByStrain(strain):
            if gn1<gn2: key = (gn1,gn2)
            else: key = (gn2,gn1)
            if key in geneProximityD and geneProximityD[key] <= proximityThreshold:
                return True
    return False
  
def rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree,rootProximate,memoD):
    '''Memoized, parsimony based calculation of cost of evolutionary rearrangement
given lfam1 and lfam2 begin at root of subtree. Assume either
proximate (nearby) or not proximate at root. This is specificed by the
argument rootProximate. Charge 1 for each rearrangment.
    '''
    memoKey = (lfam1.locusFamNum,lfam2.locusFamNum,subtree,rootProximate)
    if memoKey in memoD:
        return memoD[memoKey]
    elif subtree[1]==():
        prox=proximity(lfam1,lfam2,geneProximityD,proximityThreshold,subtree[0])
        if (prox and rootProximate) or (not prox and not rootProximate):
            # state given by rootProximate matches adjacency info in our data
            output = 0
        else:
            output = 1
    else:
        # left subtree
        left = rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree[1],rootProximate,memoD)
        chLeft = 1 + rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree[1],not rootProximate,memoD)

        # right subtree
        right = rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree[2],rootProximate,memoD)
        chRight = 1 + rcost(lfam1,lfam2,geneProximityD,proximityThreshold,subtree[2],not rootProximate,memoD)

        output = min(left,chLeft) + min(right,chRight)
        
    memoD[memoKey] = output
    return output

def rscore(li0,li1,costDiffD):
    '''Returns rearrangement score and orientation between LocusIslands
li0 and li1. Considers all four ways these locus islands could join
(since there are two locus islands each with two ends). This score is
the rcost given we start not proximate minus the rcost given we start
proximate. costDiffD is the precomputed cost diff scores between
families. orientation is an number 0-3 indicating which merge
orientation the (best) score came from.

    '''
    if li0.mrca != li1.mrca:
        return -float('inf')
    else:
        # we try combining in each of the 4 orientations (but if locus
        # island only has 1 family, then we can skip the reverse
        # orientation for that one.
        caseL = [-float('inf')]*4
        
        # case 0: last lfam in li0 vs. first lfam in li1
        caseL[0] = costDiffD[(li0.locusFamilyL[-1],li1.locusFamilyL[0])]

        # case 1: last lfam in li0 vs. last lfam in li1
        if len(li1.locusFamilyL) == 1:
            caseL[1] = caseL[0] # since first and last of li1 are same
        else:
            caseL[1] = costDiffD[(li0.locusFamilyL[-1],li1.locusFamilyL[-1])]
            
        # case 2: first lfam in li0 vs. first lfam in li1
        if len(li0.locusFamilyL) == 1:
            caseL[2] = caseL[0] # since first and last of li0 are same
        else:
            caseL[2] = costDiffD[(li0.locusFamilyL[0],li1.locusFamilyL[0])]
            
        # case 3: first lfam in li0 vs. last lfam in li1
        if len(li0.locusFamilyL) == 1 and len(li1.locusFamilyL) > 1:
            caseL[3] = caseL[1]
        elif len(li0.locusFamilyL) > 1 and len(li1.locusFamilyL) == 1:
            caseL[3] = caseL[2]
        elif len(li0.locusFamilyL) == 1 and len(li1.locusFamilyL) == 1:
            caseL[3] = caseL[0]
        else: # both longer than 1
            caseL[3] = costDiffD[(li0.locusFamilyL[0],li1.locusFamilyL[-1])]            

        score = max(caseL)
        orientation = caseL.index(score)
        
        return score,orientation

## Output

def printSummary(focalNodesL,strainNum2StrD,numIslandsAtEachNodeAtStartL,numIslandsAtEachNodeAtEndL,outputSummaryF):
    '''Print a summary of the merging saying how many islands there were
at each node in the focal clade, before and after merging.'''

    print("Number of islands per node in focal clade: ",file=outputSummaryF)
    rowL=[]
    rowL.append(['Node','Before merge','After merge'])
    rowL.append(['----','------------','-----------'])
    for mrcaNode in focalNodesL:

        rowL.append([strainNum2StrD[mrcaNode],str(numIslandsAtEachNodeAtStartL[mrcaNode]),str(numIslandsAtEachNodeAtEndL[mrcaNode])])
    
    analysis.printTable(rowL,indent=2,fileF=outputSummaryF)

    return

def writeIslands(locIslByNodeL,strainNum2StrD,islandOutFN):
    '''Write the islands to a file'''
    f=open(islandOutFN,"w")
    for branch in range(len(locIslByNodeL)):
        for island in locIslByNodeL[branch]:
            print(island.fileStr(strainNum2StrD),file=f)
    f.close()


def readIslands(islandFN,tree,strainStr2NumD):
    '''Given a file name for a islands output file, load back
recreating locIslByNodeL.'''

    locIslByNodeL=[[] for i in range(trees.nodeCount(tree))]
    
    f=open(islandFN,'r')
    while True:
        s=f.readline()
        if s == '':
            break
        gr=str2Island(s.rstrip(),strainStr2NumD)
        locIslByNodeL[gr.mrca].append(gr)
    f.close()

    return locIslByNodeL



import sys
from .Family import *
from .Island import *
from . import trees
from . import scores
from . import islands

#### Analysis functions

## general

def printTable(L,indent=0,fileF=sys.stdout):
    '''Given tabular data in a list of lists (where sublists are rows)
print nicely so columns line up. Indent is an optional number of blank spaces to put in front of each row.'''
    # get max width for each column
    colMax=[]
    for col in range(len(L[0])):
        mx=0
        for row in L:
            if len(row[col]) > mx:
                mx = len(row[col])
        colMax.append(mx)
    
    # print
    for row in L:
        for col in range(len(row)):
            row[col]=row[col]+' ' * (colMax[col]-len(row[col]))
            
    for row in L:
        printStr = " "*indent + " | ".join(row)
        print(printStr.rstrip(),file=fileF)

def matchFamilyIsland(geneInfoD,geneNames,gene2FamIslandD,searchStr):
    '''Return the island number, family number, and gene name(s)
associated with searchStr in geneInfoD. Searches for a match in all
fields of geneInfoD.'''
    # find matching gene names
    geneMatchL=[]
    for geneName in geneInfoD:
        valueT=geneInfoD[geneName]
        for value in (geneName,)+valueT:
            if type(value)==str:
                if searchStr in value:
                    geneMatchL.append(geneName)
                    break

    # get family numbers and island numbers
    outL=[]
    for geneName in geneMatchL:
        geneNum = geneNames.nameToNum(geneName)
        (locusIslandNum, famNum, locusFamNum) = gene2FamIslandD[geneNum]
        outL.append((geneName,locusIslandNum, famNum, locusFamNum))
    return outL
        
## Print scores associated with a family

def printScoreMatrix(familyNum,subtreeL,familiesO,geneNames,scoresO,scoreType,fileF):
    '''Print a matrix of scores between all the genes in a family given by
familyNum. Scores are provided by scoresO, and we're extracting the
values associated with scoreType in the edges of this graph.
    '''

    familyGeneNumsL = []
    for lfO in familiesO.getFamily(familyNum).getLocusFamilies():
        familyGeneNumsL.extend(lfO.iterGenes())
    
    rowsL = []
    geneNamesL = [geneNames.numToName(gn) for gn in familyGeneNumsL]
    rowsL.append([''] + geneNamesL)
    
    for rowi,gn1 in enumerate(familyGeneNumsL):
        row = [geneNames.numToName(familyGeneNumsL[rowi])]
        for gn2 in familyGeneNumsL:
            if scoresO.isEdgePresentByEndNodes(gn1,gn2):
                row.append(format(scoresO.getScoreByEndNodes(gn1,gn2,scoreType),".3f"))
            else:
                row.append('-')
        rowsL.append(row)

    printTable(rowsL,indent=2,fileF=fileF)

def printOutsideFamilyScores(familyNum,subtreeL,familiesO,geneNames,scoresO,fileF):
    '''Given a family, print scores for all non-family members with a
connection to genes in family. Scores are provided in the network
scoresO.
    '''

    family = familiesO.getFamily(familyNum)
    outsideGeneNumsS = family.getOutsideConnections(scoresO)
    
    rowL = []
    for familyGeneNum in family.iterGenes():
        familyGeneName = geneNames.numToName(familyGeneNum)
        for outsideGeneNum in outsideGeneNumsS:
            if scoresO.isEdgePresentByEndNodes(familyGeneNum,outsideGeneNum):
                outsideGeneName = geneNames.numToName(outsideGeneNum)
                rawSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'rawSc')
                synSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'synSc')
                coreSynSc=scoresO.getScoreByEndNodes(familyGeneNum,outsideGeneNum,'coreSynSc')
                rowL.append([familyGeneName,outsideGeneName,format(rawSc,".3f"),format(synSc,".3f"),format(coreSynSc,".3f")])

    rowL.sort(key=lambda x: x[2],reverse=True) # sort by score
    rowL.insert(0,['----------','-----------','---','---','-------'])
    rowL.insert(0,['Inside fam','Outside fam','Raw','Syn','CoreSyn'])
                
    print("Printing all scores with non-family members",file=fileF)
    printTable(rowL,indent=2,fileF=fileF)

    
## Print all islands at node

def printIslandLSummary(islandL,fileF):
    '''Given a list of locus islands in islandL (ie a list from a single node),
print a simple tabular summary indicating how many locus families they
have.
    '''
    lenL = []
    for isl in islandL:
        lenL.append(len(isl)) # len of island is num families

    # count how many times each length occurs
    lnCtD = {}
    for ln in lenL:
        if ln in lnCtD:
            lnCtD[ln] += 1
        else:
            lnCtD[ln] = 1

    # print out
    printL = []
    row = ['Num LocusFamilies in LocusIsland','Number of occurrences']
    printL.append(row)
    
    for ln,occurrences in sorted(lnCtD.items()):
        printL.append([str(ln), str(occurrences)])

    printTable(printL,indent=8,fileF=fileF)

def vPrintLocusIsland(island,subtreeL,familiesO,strainNum2StrD,geneNames,geneInfoD,fileF):
    '''Verbose print of a locus island.'''

    print("  LocusIsland",island.id,file=fileF)
    
    # get species nodes subtended by this mrca
    speciesNodesL=trees.leafList(subtreeL[island.mrca])

    # put everything in lists.
    printL=[]
    printL.append(['LocusFamily'])
    for node in speciesNodesL:
        printL[0].append(strainNum2StrD[node])
        
    for locusFamO in island.iterLocusFamilies(familiesO):
        newRow=[]
        newRow.append(str(locusFamO.locusFamNum))
        for node in speciesNodesL:
            entryL = []
            for geneNum in locusFamO.iterGenesByStrain(node):
                geneName = geneNames.numToName(geneNum)
                commonGeneName = "("+geneInfoD[geneName][0]+")" if geneInfoD[geneName][0] != '' else ''
                entryL.append(geneName + commonGeneName)
                
            if entryL == []:
                entry = ''
            else:
                entry = ",".join(entryL)

            newRow.append(entry)
        printL.append(newRow)
       
    printTable(printL,indent=4,fileF=fileF)


def vPrintLocusIslandsAtNode(islandL,subtreeL,familiesO,strainNum2StrD,geneNames,geneInfoD,fileF):
    '''Print a list of islands at a single node.'''
    print("  Summary",file=fileF)
    printIslandLSummary(islandL,fileF)
    print("  ---",file=fileF)
    for island in islandL:
        vPrintLocusIsland(island,subtreeL,familiesO,strainNum2StrD,geneNames,geneInfoD,fileF)
        print('  ---',file=fileF)

def vPrintAllLocusIslands(islandByNodeL,tree,rootFocalClade,subtreeL,familiesO,strainStr2NumD,strainNum2StrD,geneNames,geneInfoD,fileF):
    '''Loop over all nodes in tree, printing islands at each. '''
    rootFocalCladeNum = strainStr2NumD[rootFocalClade]
    focalTree = trees.subtree(tree,rootFocalCladeNum)
    for node in trees.nodeList(focalTree):
        nodeStr = strainNum2StrD[node]
        print('########################### ',"Locus Islands at node",nodeStr,file=fileF)
        print('',file=fileF)
        vPrintLocusIslandsAtNode(islandByNodeL[node],subtreeL,familiesO,strainNum2StrD,geneNames,geneInfoD,fileF)

def createGene2FamIslandD(islandL,familiesO):
    '''Creates a dictionary keyed by gene number which has the
LocusIsland, Family and LocusFamily for each gene.'''
    D = {}
    for islandsAtNodeL in islandL:
        for locusIslandO in islandsAtNodeL:
            for locusFamO in locusIslandO.iterLocusFamilies(familiesO):
                for gene in locusFamO.iterGenes():
                    D[gene] = (locusIslandO.id,locusFamO.famNum,locusFamO.locusFamNum)

    return D

## Print species files with all the genes, grouped by contig

def printSpeciesContigs(geneOrderT,fileStemStr,fileExtensionStr,geneNames,gene2FamIslandD,geneInfoD,familiesO,strainNum2StrD):
    '''This function produces a set of species specific genome
files. These contain all the genes in a strain laid out in the order
they occur on the contigs. Each gene entry includes LocusIsland and LocusFamily
information, as well as a brief description of the gene's function.'''

    for strainNum in range(len(geneOrderT)):
        strainStr = strainNum2StrD[strainNum]
        contigT = geneOrderT[strainNum]
        if contigT != None:
            # it's a tip
            fileF=open(fileStemStr+'-'+strainStr+fileExtensionStr,'w')
            for contig in contigT:
                print("########### Contig",file=fileF)
                printGenes(contig,geneNames,gene2FamIslandD,geneInfoD,[],familiesO,strainNum2StrD,fileF)

            fileF.close()
    
## Print neighborhood of an island

def printIslandNeighb(islandNum,synWSize,subtreeL,islandByNodeL,familiesO,geneOrderT,gene2FamIslandD,geneInfoD,geneNames,strainNum2StrD,fileF):
    '''Print the neighborhood of an island. We include the genes in the island and synWSize/2 genes in either direction.'''

    print("  LocusIsland:",islandNum,file=fileF)
    
    genesInEitherDirec = int(synWSize/2)

    # get the island object for this islandNum
    for listOfIslands in islandByNodeL:
        _,island = islands.searchLocIslandsByID(listOfIslands,islandNum)
        if island != None: break

    if island == None:
        raise ValueError("LocusIsland "+str(islandNum)+" not found.")
        
    mrca = island.mrca
    print("  mrca:",strainNum2StrD[mrca],file=fileF)

    leavesL=trees.leafList(subtreeL[mrca])

    for strainNum in leavesL:

        print("  In",strainNum2StrD[strainNum],end=' ',file=fileF)

        islandGenesInStrainL = getIslandGenesInStrain(island,strainNum,familiesO)

        if islandGenesInStrainL == []:
            print("the island is not found.",file=fileF)
        else:

            neighbGenesL,firstIslandGene,lastIslandGene=getNeighborhoodGenes(strainNum,geneOrderT,islandGenesInStrainL,genesInEitherDirec)

            # print coordinates of island in this strain
            chrom=geneInfoD[geneNames.numToName(islandGenesInStrainL[0])][3]
            startPos = geneInfoD[geneNames.numToName(firstIslandGene)][4]
            endPos = geneInfoD[geneNames.numToName(lastIslandGene)][5]

            print("(Coordinates",chrom+":"+str(startPos)+"-"+str(endPos)+")",file=fileF)

            printGenes(neighbGenesL,geneNames,gene2FamIslandD,geneInfoD,islandGenesInStrainL,familiesO,strainNum2StrD,fileF)


def getIslandGenesInStrain(island,strainNum,familiesO):
    '''Given an island, a strain number, and our families object, return
all the genes in the island for that strain.'''
    genesL=[]
    for locFam in island.iterLocusFamilies(familiesO):
        genesL.extend(locFam.iterGenesByStrain(strainNum))
    return genesL

def getNeighborhoodGenes(strainNum,geneOrderT,islandGenesInStrainL,genesInEitherDirec):
    ''''''
    neighbGenesL=[]
    for contig in geneOrderT[strainNum]:
        try:
            # get index of all of these. We're assuming they're on
            # the same contig and in the same area.
            indL=[contig.index(gene) for gene in islandGenesInStrainL]
            maxInd = max(indL)
            minInd = min(indL)
            end = maxInd + genesInEitherDirec +1
            st = minInd-genesInEitherDirec if minInd-genesInEitherDirec>0 else 0 # st can't be less than 0
            neighbGenesL=contig[st:end]

            # get gene numbers of first and last genes in island
            firstGene = contig[minInd]
            lastGene = contig[maxInd]
            
            return neighbGenesL,firstGene,lastGene
        except ValueError:
            continue

def printGenes(neighbGenesL,geneNames,gene2FamIslandD,geneInfoD,islandGenesInStrainL,familiesO,strainNum2StrD,fileF):
    '''Given a list of contiguous genes, print them out nicely with
information on locus family and locus island etc. Put * next to any
that are in islandGenesInStrainL.'''
            
    # now print the neighbors
    rowsL=[]
    for gene in neighbGenesL:
        geneName=geneNames.numToName(gene)
        locIslNum,famNum,locFamNum = gene2FamIslandD[gene]
        lfMrca = familiesO.getLocusFamily(locFamNum).lfMrca

        
        if geneName in geneInfoD:
            descrip = geneInfoD[geneName][2]
        else:
            descrip = ''

        # mark genes in the island with a *
        if gene in islandGenesInStrainL:
            geneName = '* '+geneName
        else:
            geneName = '  '+geneName

        infoL = [geneName,"locIsl:"+str(locIslNum),"fam:"+str(famNum),"locFam:"+str(locFamNum),"locFamMRCA:"+strainNum2StrD[lfMrca],descrip]

        rowsL.append(infoL)

    printTable(rowsL,indent=4,fileF=fileF)

# support functions for printCoreNonCoreByNode

def createFamilyByNodeL(geneOrderT,gene2LocFamD):
    '''Create a list where index is node, and where the value is all the
families present in that strain. Internal nodes are None, tips are a set of
families.'''

    # an index for each node. internal nodes get none, just like in geneOrderT
    familyByNodeL=[set() if x!=None else None for x in geneOrderT]

    for i in range(len(geneOrderT)):
        contigT = geneOrderT[i]
        if contigT != None:
            for contig in contigT:
                for geneNum in contig:
                    fam=gene2LocFamD[geneNum]
                    familyByNodeL[i].add(fam)
    return familyByNodeL

def coreNonCoreCtAtNode(tree,node,familyByNodeL,familiesO):
    '''Given a tree and a node, first get all the families present in
descendant species. Then figure out which of these families are
non-core (their mrca is located below node) and which are core (mrca
is at node or above). Return count of non-core and core.'''

    subtree = trees.subtree(tree,node)
    nonCoreMrcaL = trees.nodeList(subtree[1]) + trees.nodeList(subtree[2])
    
    # get set of all families with members in descendant species of this node
    decFamS = set()
    for leaf in trees.leafList(subtree):
        decFamS.update(familyByNodeL[leaf])

    # figure out which are core, non-core
    coreCt=0
    nonCoreCt=0
    totCt = len(decFamS)
    for fam in decFamS:
        if familyL[fam].mrca in nonCoreMrcaL:
            nonCoreCt += 1
        else:
            coreCt += 1
            
    return nonCoreCt,coreCt



# Functions for loading genes and gene order
import sys
from . import fasta
from . import trees

def loadProt(protFnL):
    '''Given a list of file names of fasta files with the gene name as
header, load the sequences and store in a dictionary keyed by protein
name.
    '''
    seqD={}
    for fn in protFnL:
        for header,seq in fasta.load(fn):
            gn = header.split()[0][1:]
            seqD[gn]=seq
    return seqD

class geneNames:
    def __init__(self, geneOrderFN,strainNameToNumD,strainNumToNameD):
        '''Create a geneName object with lists of genes and methods to interconvert.'''

        self.strainNameToNumD = strainNameToNumD
        self.strainNumToNameD = strainNumToNameD
        
        # get lists of genes
        names=[]
        num=0
        geneNumToStrainNumD={}
        f = open(geneOrderFN,'r')
        while True:
            s = f.readline()
            if s == '':
                break
            L=s.split()
            strain = L[0]
        
            for i in range(1,len(L)): 
                geneName=L[i]
                names.append(geneName)
                geneNumToStrainNumD[num] = strainNameToNumD[strain]

                num+=1

        f.close()

        namesS=set(names)
        if len(namesS) < len(names):
            raise ValueError("There appear to be redudancies in the gene order file.")

        self.geneNumToStrainNumD = geneNumToStrainNumD
        self.names=tuple(names)
        self.nums=tuple(range(len(names))) # these will be the gene numbers for the genes
        
        # create dictionaries for interconverting
        geneNameToNumD={}
        geneNumToNameD={}
        for i in range(len(self.names)):
            geneName=self.names[i]
            num = self.nums[i]
        
            geneNameToNumD[geneName]=num
            geneNumToNameD[num]=geneName

        self.geneNameToNumD = geneNameToNumD
        self.geneNumToNameD = geneNumToNameD


    def nameToNum(self,geneName):
        return self.geneNameToNumD[geneName]

    def numToName(self,geneNumber):
        return self.geneNumToNameD[geneNumber]

    def numToStrainNum(self,geneNumber):
        return self.geneNumToStrainNumD[geneNumber]

    def nameToStrainNum(self,geneName):
        return self.geneNumToStrainNumD[self.nameToNum(geneName)]

    def numToStrainName(self,geneNumber):
        return self.strainNumToNameD[self.geneNumToStrainNumD[geneNumber]]

    def nameToStrainName(self,geneName):
        return self.strainNumToNameD[self.geneNumToStrainNumD[self.nameToNum(geneName)]]

    def isSameStrain(self,geneNum1,geneNum2):
        '''Tests if two genes, given in numerical form, are in the same
strain. Returns boolean.
        '''
        return self.numToStrainNum(geneNum1) == self.numToStrainNum(geneNum2)
        
    def __repr__(self):
        return "<geneName object with "+str(len(self.names))+" genes.>"
    
def readGeneInfoD(geneInfoFN):
    '''Read gene info from file, returning a dict keyed by gene name with
information such as description, start position and so on.
    '''
    geneInfoD = {}
    f = open(geneInfoFN,'r')
    while True:
        s = f.readline()
        if s == '':
            break
        geneName,commonName,locusTag,descrip,chrom,start,end,strand=s.rstrip().split('\t')
        geneInfoD[geneName]=(commonName,locusTag,descrip,chrom,start,end,strand)
    f.close()
    return geneInfoD

def getProximityInWindow(geneWinT,geneProximityD):
    '''Given a window of genes, calculate the distances between first gene
and the rest (in number of genes) and store in geneProximityD. Updates
geneProximityD in place, returning None.
    '''
    gnA=geneWinT[0]
    for i in range(1,len(geneWinT)):
        gnB=geneWinT[i]

        if gnA<gnB: # always put lower gene number first
            geneProximityD[(gnA,gnB)]=i
        else:
            geneProximityD[(gnB,gnA)]=i

def createGeneProximityD(geneOrderT,geneProximityForGroup):
    '''Go though a gene order tuple pulling out pairs of genes that are
within geneProximityForGroup genes of each other. Store in a dict keyed by
gene pair, with value equal to the distance between the two genes,
measured in number of genes.'''
    geneProximityD = {}
    for contigT in geneOrderT:
        if contigT != None:
            # internal nodes are None, having no genes to be adjacent
            for geneNumT in contigT:
                for i in range(len(geneNumT)):
                    # slide over genes w/win geneProximityForGroup+1
                    # wide.
                    getProximityInWindow(geneNumT[i:i+(geneProximityForGroup+1)],geneProximityD)
                    
    return geneProximityD

def createGeneOrderTs(geneOrderFN,geneNames,subtreeL,strainStr2NumD):
    '''Go though gene order file and get orderings into a set of
tuples. Returns a tuple whose index is strain number, and the value at
that index is a set of tuples representing the contigs.'''
    f = open(geneOrderFN,'r')
    geneOrderL=[None for x in range(trees.nodeCount(subtreeL[-1]))] # an index for each node
    while True:
        s = f.readline()
        if s == '':
            break
        s=s.rstrip()
        # note, our gene order format has contigs separated by \t, and
        # genes within them separated by a space character.
        L=s.split('\t')
        strain = L[0]
        contigL=[]
        for contig in L[1:]:
            geneNumT=tuple((geneNames.nameToNum(g) for g in contig.split(' ')))
            contigL.append(geneNumT)
            
        geneOrderL[strainStr2NumD[strain]]=tuple(contigL)
    f.close()
    return tuple(geneOrderL)




import sys, os, subprocess, glob
from . import trees
from multiprocessing import Pool
from Bio import Phylo

def getDbFileL(fastaFilePath,treeFN):
    '''Obtains and returns a list of all fasta files that should be run
through blast. If there is a valid tree file present, then this list
should contain only those fastas that correspond to species in the
tree. If there is not a valid tree file, then the list will contain
all fasta files present in the fasta directory. The reason we don't
want to assume there is a tree file is that sometimes users will need
to run blast before making that.
    '''
    # get list of all fastas in directory
    dbFileL=glob.glob(fastaFilePath)

    # load tree if present
    try:
        newDbFileL=[]
        tree,strainStr2NumD,strainNum2StrD = trees.readTree(treeFN)
        leavesL=[strainNum2StrD[strainNum] for strainNum in trees.leafList(tree)]

        for dbFile in dbFileL:
            dbStem = os.path.split(dbFile)[-1] # strain + extension
            dbStem = os.path.splitext(dbStem)[0] # strain only
            if dbStem in leavesL:
                # this file name is in the tree, include it
                newDbFileL.append(dbFile)

        dbFileL = newDbFileL
    except ( TypeError, FileNotFoundError, Phylo.NewickIO.NewickError ):
        print("Since a properly formated tree file was not provided, we can't use that to determine which fasta files to blast. We will therefore blast all fasta files in fastaFilePath.",file=sys.stderr)

    return dbFileL

def formatDb(dbFileL,blastExecutDirPath):
    '''Format fasta files in fastaDir for blast by calling the formatdb
executable.'''
    for dbFileName in dbFileL:
        makeblastdbExecutable = os.path.join(blastExecutDirPath,'makeblastdb')
        subprocess.call([makeblastdbExecutable, '-dbtype' ,'prot', '-in', dbFileName],stdout=subprocess.PIPE)
    return

def makeBlastClineList(dbFileL_1, dbFileL_2,paramD):
    '''Create a list of lists, where the sublists have the command line
needed to run blastp on a pair of databases.'''

    fastaFilePath = paramD['fastaFilePath']
    blastFilePath = paramD['blastFilePath']
    blastExecutDirPath = paramD['blastExecutDirPath']

    # catch blast clines from old params files which don't have
    # trailing whitespace (can get rid of this eventually).
    if paramD['blastCLine'][-1] != ' ':
        blastCLine = paramD['blastCLine'] + ' ' + str(paramD['evalueThresh'])
    else:
        blastCLine = paramD['blastCLine'] + str(paramD['evalueThresh'])
    
    # get a tuple of the blastp command line args for use below
    blastCLineL = blastCLine.split()
    blastCLineL[0] = os.path.join(blastExecutDirPath,blastCLineL[0])
    blastCLineT = tuple(blastCLineL)

    # get blast file dir and extension
    splitT = os.path.split(blastFilePath)
    blastDir,rest = splitT
    blastExtension = rest.split("*")[-1]

    clineL=[]
    for query in dbFileL_1:
        for db in dbFileL_2:

            qstem = os.path.split(query)[-1]
            qstem = os.path.splitext(qstem)[0]

            dbstem = os.path.split(db)[-1]
            dbstem = os.path.splitext(dbstem)[0]

            outFN = os.path.join( blastDir, qstem + '_-VS-_' + dbstem + blastExtension )
            L = list(blastCLineT) + ['-query',query,'-db',db,'-out',outFN]
            clineL.append(L)

    return clineL

def subprocessWrapper(cline):
    '''A wraper within which we call subprocess. Becasue we're having
blast write to file, we dump std_out. return std_err.'''
    pipes=subprocess.Popen(cline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = pipes.communicate()
    return stderr

def runBlast(dbFileL_1,dbFileL_2,paramD):
    '''Run blast comparing every database in dbFileL_1 against every
database in dbFileL_2.
    '''

    # format the databases
    uniqueDbL=list(set(dbFileL_1+dbFileL_2))
    formatDb(uniqueDbL,paramD['blastExecutDirPath'])
    
    # if directory for blast doesn't exist yet, make it
    blastDir = os.path.split(paramD['blastFilePath'])[0]
    if glob.glob(blastDir)==[]:
        os.mkdir(blastDir)

    clineL =  makeBlastClineList(dbFileL_1,dbFileL_2,paramD)

    p=Pool(paramD['numThreads'])
    stderrL = p.map(subprocessWrapper, clineL)




import numpy,glob,os,struct,statistics,ctypes
from multiprocessing import RawArray


class Score:

    def __init__(self):
        '''Create an object for storing scores.'''
        
        self.endNodesToEdgeD = {}
        self.numEdges = 0 # to start out
        self.strainPairScoreLocationD = {}
        self.scoreD = {}

    def initializeDataAttributes(self,blastFilePath,geneNames,strainStr2NumD):
        '''This method takes a new, empty object and fills the data attributes
by reading through blast files to identify pairs of genes with
edges. Also creates the array for storing raw scores, initialized to
0. Arrays for other score types will be created later.
        '''
        self.fillEndNodesToEdgeD(blastFilePath,geneNames,strainStr2NumD)
        self.numEdges=len(self.endNodesToEdgeD)
        self.initializeScoreArray('rawSc')

    def fillEndNodesToEdgeD(self,blastFilePath,geneNames,strainStr2NumD):
        '''Run through blast files, finding all pairs of genes with signicant
similarity. Use these to fill endNodesToEdgeD. Also keep track of the
edge numbers associated with particular strain pairs and save in
strainPairScoreLocationD.
        '''

        # get blast files organized by strain pairs
        
        blastFnByPairD = self.getBlastFnByPairD(blastFilePath,strainStr2NumD)

        edgeNum=0
        for strainPair in blastFnByPairD:
            pairT = blastFnByPairD[strainPair]
            
            strainPairSt = edgeNum # scores for this strain pair will
                                   # begin at edgeNum

            # pairT may have two blast files (e.g for strain 1 vs 2
            # and also for 2 vs 1), or may only have one (e.g. for
            # strain 1 vs strain 1).
            for fn in pairT:
            
                f = open(fn,'r')
                while True:
                    s = f.readline()
                    if s=='':
                        break
                    L = s.split('\t')
                    if len(L) != 12: # we only want lines with 12 columns
                        continue

                    g1 = geneNames.nameToNum(L[0])
                    g2 = geneNames.nameToNum(L[1])

                    # because we have blast files going both ways
                    # (e.g. strain 1 vs strain 2 and also strain 2 vs
                    # strain 1), we check if we've added this
                    # particular gene pair already. If so, no need to
                    # do so again.
                    
                    # make sure g1 is the lower gene number (we do this so
                    # we only store in one orientation in dict).
                    if g1 > g2: g2,g1 = g1,g2

                    if (g1,g2) not in self.endNodesToEdgeD:
                        # we haven't got it yet, add it
                        self.endNodesToEdgeD[(g1,g2)] = edgeNum
                        edgeNum += 1
                f.close()
                
            strainPairEnd = edgeNum

            # Store the range of edge numbers for this
            # strainPair. Keys for this dict are tuples of train
            # number, e.g. (1,2) Always lower number first.
            self.strainPairScoreLocationD[strainPair] = (strainPairSt,strainPairEnd)
            

    def getBlastFnByPairD(self,blastFilePath,strainStr2NumD):
        '''Get the set of blast files and organize by the pair of strains
compared. Returns a dict keyed by tuple of strain number
(e.g. (1,2)). The value is a list with all the blast files comparing
two strains. For example one list might contain
['blast/E_albertii-E_coli_K12.out',
'blast/E_coli_K12-E_albertii.out']. For cases where a strain is
compared against itself then we have only one file name in the list.
        '''
        blastFnL=glob.glob(blastFilePath)

        blastFnByPairD ={}
        
        for fileStr in blastFnL:
            strain1,strain2 = os.path.splitext(os.path.split(fileStr)[-1])[0].split('_-VS-_')
            strainNum1 = strainStr2NumD[strain1]
            strainNum2 = strainStr2NumD[strain2]            
            key = tuple(sorted([strainNum1,strainNum2]))
            if key in blastFnByPairD:
                # we've already looked at the one that had them in the other order
                blastFnByPairD[key].append(fileStr)
            else:
                blastFnByPairD[key] = [fileStr]
            
        return blastFnByPairD
                
    def initializeScoreArray(self,scoreType):
        '''Create array for storing scores.'''
        self.scoreD[scoreType] = numpy.zeros(self.numEdges,dtype=ctypes.c_double)

    def getStrainPairs(self):
        '''Return all the strain pairs associated with this scores object as a
list of tuples. Each tuple will be two strains in numeric form.'''
        return sorted(self.strainPairScoreLocationD.keys())
        
    def isEdgePresentByEndNodes(self,g1,g2):
        '''See if g1 and g2 have an edge in our data structure. Return
boolean.'''
        if g1 > g2: g2,g1 = g1,g2
        if (g1,g2) in self.endNodesToEdgeD: return True
        else: return False
        
    def endNodesToEdge(self,g1,g2):
        '''Given two genes, return the number of the edge between them. If
there isn't any, return None.'''
        edge = self.endNodesToEdgeD[(g1,g2)]
        return edge

    def addScoreByEdge(self,edge,sc,scoreType):
        '''Given a score on an edge, store it in the score array corresponding
to scoreType.
        '''
        self.scoreD[scoreType][edge] = sc
        
    def addScoreByEndNodes(self,g1,g2,sc,scoreType):
        '''Given a score between two genes store it in the score array
corresponding to scoreType.'''

        if g1 > g2: g2,g1 = g1,g2 # make sure g1 is lower gene num

        if self.isEdgePresentByEndNodes(g1,g2):
            # we can only add scores for those edges that already exist
            edge = self.endNodesToEdge(g1,g2)
            self.addScoreByEdge(edge,sc,scoreType)
        # if the edge isn't already present, do nothing.
            
    def getScoreByEdge(self,edge,scoreType):
        '''Given an edge get the score corresponding to scoreType.'''
        return self.scoreD[scoreType][edge]

    def getScoreByEndNodes(self,g1,g2,scoreType):
        '''Given two genes get the score corresponding to scoreType.'''

        if g1 > g2: g2,g1 = g1,g2 # make sure g1 is lower gene num
        edge = self.endNodesToEdge(g1,g2)
        return self.getScoreByEdge(edge,scoreType)

    def iterateEdges(self):
        '''Returns an iterator which goes over edges by edge number.'''
        return range(self.numEdges)

    def iterateEdgesByEndNodes(self):
        '''Returns an iterator which goes over edges by end nodes.'''
        return self.endNodesToEdgeD.keys()

    def iterateScoreByStrainPair(self,strainPair,scoreType):
        '''Returns an iterator which gives all scores of type scoreType associated with a
particular strainPair.'''
        stInd,endInd = self.strainPairScoreLocationD[strainPair]
        for edge in range(stInd,endInd):
            yield self.getScoreByEdge(edge,scoreType)

    def __eq__(self,other):
        '''Determine if other has the same values as self. Useful mainly for debugging.'''

        if self.numEdges != other.numEdges:
            return False

        if sorted(self.endNodesToEdgeD.items()) != sorted(other.endNodesToEdgeD.items()):
            return False

        if sorted(self.strainPairScoreLocationD.items()) != sorted(other.strainPairScoreLocationD.items()):
            return False

        if sorted(self.scoreD.keys()) != sorted(other.scoreD.keys()):
            return False
        
        for scoreType in self.scoreD.keys():
            if not all(self.scoreD[scoreType] == other.scoreD[scoreType]):
                return False

        # we made it, they're the same
        return True
        
    def writeScoresText(self,geneNames,scoreTypeL,scoresFN):
        '''Save all edges (pairs of genes) to a tab delimited text file with a
    header line.
        '''
        # open file
        f=open(scoresFN,'w')

        # Specify the index ranges for different strain pairs
        f.write("# Index range of scores for different strain pairs: strain1 strain2 stInd endInd.\n")
        for key in self.strainPairScoreLocationD:
            strain1,strain2 = key
            stInd,endInd = self.strainPairScoreLocationD[key]
            f.write("\t".join(map(str,[strain1,strain2,stInd,endInd]))+"\n")
            
        # write header
        f.write("# Scores: "+"\t".join(['gene1','gene2','edge']+scoreTypeL)+'\n')

        for gene1Num,gene2Num in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[(gene1Num,gene2Num)]
            outStrL=[]
            outStrL.append(geneNames.numToName(gene1Num))
            outStrL.append(geneNames.numToName(gene2Num))
            outStrL.append(str(edge))
            for scoreType in scoreTypeL:
                outStrL.append(format(self.getScoreByEndNodes(gene1Num,gene2Num,scoreType),".6f"))

            f.write("\t".join(outStrL)+'\n')

        f.close()

    def readScoresText(geneNames,scoreTypeL,scoresFN):
        '''Read scores from a text file of scores and use to create Score
        object
        '''

        scoresO = Score()
        f = open(scoresFN,'r')

        # load the index ranges for different strain pairs
        f.readline() # skip first header/separator line
        
        maxEndInd = 0 # the biggest endInd we've seen so far
        while True:
            s = f.readline()
            if s[0] == '#':
                # we've hit the separator. scores begin on next line
                break
            lineL=s.split('\t')
            strain1 = int(lineL[0])
            strain2 = int(lineL[1])
            stInd = int(lineL[2])
            endInd = int(lineL[3])

            if endInd > maxEndInd:
                maxEndInd = endInd
            
            scoresO.strainPairScoreLocationD[(strain1,strain2)] = (stInd,endInd)

        # now we move to loading scores. The number of edges will be maxEndInd

        # Set up the arrays
        scoresO.numEdges = maxEndInd
        for scoreType in scoreTypeL:
            scoresO.initializeScoreArray(scoreType)

        separaterLineL = s.split()
        scoreTypeL=separaterLineL[5:] # the types of scores are listed in this separator line
        
        while True:
            s = f.readline()
            if s == '':
                break
            lineL=s.split('\t')

            g1 = geneNames.nameToNum(lineL[0])
            g2 = geneNames.nameToNum(lineL[1])
            edge = int(lineL[2])
            scoresO.endNodesToEdgeD[(g1,g2)] = edge

            scoreL=lineL[3:]
            for i,sc in enumerate(scoreL):
                scoresO.addScoreByEdge(edge,float(sc),scoreTypeL[i])

        f.close()
        return scoresO

    def writeScoresBinary(self,scoreTypeL,scoresFN):
        '''Write scores to scoresFN as a binary file.'''

        f=open(scoresFN,'wb')

        # first 8 bytes are int, telling us how many edges
        f.write(self.numEdges.to_bytes(8,'little'))

        # next 8 bytes are int telling us how many strain pairs there are
        numStrainPairs = len(self.strainPairScoreLocationD)
        f.write(numStrainPairs.to_bytes(8,'little'))

        # write a block of bytes for each strain pair location
        for key in self.strainPairScoreLocationD:
            strain1,strain2 = key
            stInd,endInd = self.strainPairScoreLocationD[key]
            f.write(strain1.to_bytes(8,'little'))
            f.write(strain2.to_bytes(8,'little'))
            f.write(stInd.to_bytes(8,'little'))
            f.write(endInd.to_bytes(8,'little'))
            
        # write a block of bytes for each edge.
        # g1 g2 edge
        for g1,g2 in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[(g1,g2)]
            f.write(g1.to_bytes(8,'little'))
            f.write(g2.to_bytes(8,'little'))
            f.write(edge.to_bytes(8,'little'))
            for scoreType in scoreTypeL:
                val = float(self.getScoreByEndNodes(g1,g2,scoreType))
                f.write(struct.pack('<d',val))
        
        f.close()
        
    def readScoresBinary(scoreTypeL,scoresFN):
        '''Read scores from a binary files.'''

        scoresO=Score()
        f=open(scoresFN,'rb')

        # read first 8 bytes to get numEdges
        b = f.read(8)
        scoresO.numEdges = int.from_bytes(b,'little')

        # read next 8 bytes to get the number of strain pairs
        b = f.read(8)
        numStrainPairs = int.from_bytes(b,'little')

        # read in the index locations where scores from different
        # strain pairs will be found
        for i in range(numStrainPairs):
            strain1 = int.from_bytes(f.read(8),'little')
            strain2 = int.from_bytes(f.read(8),'little')
            stInd = int.from_bytes(f.read(8),'little')
            endInd = int.from_bytes(f.read(8),'little')
            scoresO.strainPairScoreLocationD[(strain1,strain2)] = (stInd,endInd)
            
        # initialize score arrays
        for scoreType in scoreTypeL:
            scoresO.initializeScoreArray(scoreType)
        
        # loop over blocks of bytes where each block is an edge
        for i in range(scoresO.numEdges):
            g1 = int.from_bytes(f.read(8),'little')
            g2 = int.from_bytes(f.read(8),'little')
            edge = int.from_bytes(f.read(8),'little')
            
            scoresO.endNodesToEdgeD[(g1,g2)] = edge

            # fill array
            for scoreType in scoreTypeL:
                scoresO.scoreD[scoreType][edge] = struct.unpack('<d',f.read(8))[0]
            
        f.close()
        return scoresO
    
    ## Below are functions that create various optional
    ## attributes. These attributes are not saved in the file format,
    ## and so must be calculated each time we intend to use them.

    def createNodeConnectL(self,geneNames):
        '''Create an attribute nodeConnectL. Index in this list corresponds to
gene. Value at that index is a list of the genes which a given gene
connects to. This attribute is not saved in our file formats. It must
be recalculated before it will be used (e.g. in family formation).'''

        self.nodeConnectL = [[] for gn in geneNames.nums]

        # loop over endNodesToEdgeD populating nodeConnectL
        for gn1,gn2 in self.endNodesToEdgeD:
            self.nodeConnectL[gn1].append(gn2)
            self.nodeConnectL[gn2].append(gn1)

    def getConnectionsGene(self,gene):
        '''Return a list containing all the genes connected to gene.'''
        return self.nodeConnectL[gene]

    def createNodeEdgeL(self,geneNames):
        '''Create an attribute nodeEdgeL. Index in this list corresponds to
gene. Value at that index is a list of the edges which a give gene
connects to. This attribute is not saved in our file formats. It must
be recalculated before it will be used.'''

        self.nodeEdgeL = [[] for gn in geneNames.nums]

        # loop over endNodesToEdgeD populating nodeEdgeL
        for gn1,gn2 in self.endNodesToEdgeD:
            edge = self.endNodesToEdge(gn1,gn2)
            self.nodeEdgeL[gn1].append(edge)
            self.nodeEdgeL[gn2].append(edge)

    def getConnectionsEdge(self,gene):
        '''Return a list containing all the edges connected to gene.'''
        return self.nodeEdgeL[gene]

    def createEdgeToEndNodeL(self):
        '''Create an attribute edgeToEndNodeL where the index is edge number,
and the value is a tuple giving the two genes on either end of the
edge. This attribute is not saved in our file formats. It must be
recalculated before it will be used.
        '''
        self.edgeToEndNodeL = [None for x in range(len(self.endNodesToEdgeD))]
        
        for endPairT in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[endPairT]
            self.edgeToEndNodeL[edge] = endPairT

    def getEndNodesByEdge(self,edge):
        '''Given and edge, return the numbers for the two genes on either end.'''
        return self.edgeToEndNodeL[edge]

    def createAabrhScoreSummaryD(self,strainNumsL,aabrhL,geneNames):
        '''Given raw scores and set of all around best reciprocal hits,
    calculates the mean and standard deviation of scores and stores in a
    dictionary.'''

        strainNumsL.sort()
        
        # create dictionary keyed by species pair, (representing an upper
        # triangular matrix)
        spScoreD={}
        for i in range(len(strainNumsL)-1):
            strain1 = strainNumsL[i]
            for j in range(i+1,len(strainNumsL)):
                strain2 = strainNumsL[j]
                spScoreD[(strain1,strain2)]=[]

        # loop through aabrhL and populate
        for orthoT in aabrhL:
            spScoreD = self.addPairwiseScores(spScoreD,orthoT,geneNames)

        # get mean and standard deviation
        self.scoreSummaryD = {}
        for sp1,sp2 in spScoreD:
            mean = statistics.mean(spScoreD[(sp1,sp2)])
            std = statistics.stdev(spScoreD[(sp1,sp2)])
            self.scoreSummaryD[(sp1,sp2)] = (mean,std)
            self.scoreSummaryD[(sp2,sp1)] = (mean,std)

    def addPairwiseScores(self,spScoreD,orthoT,geneNames):
        '''Given a dictionary for storing pairwise scores, and ortholog set in
    orthoT, and a network of scores, scoresO, pull out all species pairs, and
    add score for each in appropriate place in spScoreD.'''

        for i in range(len(orthoT)-1):
            gene1 = orthoT[i]
            geneNum1=geneNames.nameToNum(gene1)
            sp1 = geneNames.nameToStrainNum(gene1)
            for j in range(i+1,len(orthoT)):
                gene2 = orthoT[j]
                geneNum2=geneNames.nameToNum(gene2)
                sp2 = geneNames.nameToStrainNum(gene2)
                sc = self.getScoreByEndNodes(geneNum1,geneNum2,'rawSc')
                key = tuple(sorted([sp1,sp2]))
                spScoreD[key].append(sc)
        return spScoreD

    
class sharedScore:

    def __init__(self):
        '''Create an object for storing scores which will efficiently share
memory across processes.'''

        self.scoreD = {}
    
    def createArrays(self,scoresO,paramD):
        '''Creates a group of multiprocessing raw arrays for use in a
sharedScores object.
        '''

        hashArrayScaleFactor = paramD['hashArrayScaleFactor']
        
        numEdges = scoresO.numEdges
        self.hashArrayLen = hashArrayScaleFactor * numEdges
        
        ## Preliminary processing
        tempHashL=[[] for i in range(self.hashArrayLen)]
        maxGnNum=0 # get maximum numeric value used to represent a gene
        for gn1,gn2 in scoresO.iterateEdgesByEndNodes():
            edge = scoresO.endNodesToEdge(gn1,gn2)
            hs = self.hashByGenePair(gn1,gn2)
            tempHashL[hs].append((gn1,gn2,edge))
            
            if max(gn1,gn2) > maxGnNum:
                maxGnNum = max(gn1,gn2)
            
        # get length for collisionAr
        collisionArLen = 0
        maxEdgesPerHash = 0
        for L in tempHashL:
            L.sort() # sort on gn1
            if len(L) > 0:
                # add extra 1 because first block of this region will
                # hold the length of the region
                collisionArLen += len(L)
            if len(L) > maxEdgesPerHash:
                maxEdgesPerHash = len(L)
        self.collisionArLen = collisionArLen
                
        ## some checks
        
        # make sure data isn't too big. Assumes hash and gene arrays are c_uint32 (32
        # bit, unsigned)
        maxPossibleGeneOrEdgeNumber = 2**32-1
        if numEdges-1 > maxPossibleGeneOrEdgeNumber:
            raise ValueError("Error creating sharedScore object. Data set has too many score pairs for the data type we're using in our shared score arrays (c_uint32).")

        if maxEdgesPerHash > 2**16:
            raise ValueError("Error creating sharedScore object. The maximum number of score pairs per hash exceeds what lenOfRegionA can hold.")

        if maxGnNum > maxPossibleGeneOrEdgeNumber:
            raise ValueError("Error creating sharedScore object. At least one gene has been given a number too large for the array data type we are using.")


        ## create arrays

        rawScoreAr = RawArray(ctypes.c_double, numEdges)
        hasEdgeAr = RawArray(ctypes.c_bool, self.hashArrayLen) # tells if any edges at given hash
        hashAr = RawArray(ctypes.c_uint32, self.hashArrayLen) # stores index to col Ars
        lenOfRegionAr = RawArray(ctypes.c_uint16, self.hashArrayLen) # stores length of region holding stuff from a given hash
        colGn1Ar = RawArray(ctypes.c_uint32, self.collisionArLen)
        colGn2Ar = RawArray(ctypes.c_uint32, self.collisionArLen)
        colEdgeAr = RawArray(ctypes.c_uint32, self.collisionArLen)
        
        ## fill rawScoreAr
        for edge in scoresO.iterateEdges():
            sc = scoresO.getScoreByEdge(edge,'rawSc')
            rawScoreAr[edge] = sc

        ## fill hash and collision Ars
        colAr_i = 0 # index into col Ars, e.g. colGn1Ar
        for hs in range(len(tempHashL)):
            if tempHashL[hs] == []:
                hasEdgeAr[hs] = False
            else:
                hasEdgeAr[hs] = True
                lenOfRegionAr[hs] = len(tempHashL[hs])
                hashAr[hs] = colAr_i
                for gn1,gn2,edge in tempHashL[hs]:
                    colGn1Ar[colAr_i] = gn1
                    colGn2Ar[colAr_i] = gn2
                    colEdgeAr[colAr_i] = edge
                    colAr_i += 1

        self.insertArrays(rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,self.hashArrayLen)

    def hashByGenePair(self,gn1,gn2):
        '''Hash to an int, modulo array len.'''
        # python hash works better than cantor pairing func
        return hash((gn1,gn2)) % self.hashArrayLen

    def insertArrays(self,rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,hashArrayLen):
        '''Attach the input arrays to self.'''
        self.scoreD['rawSc'] = rawScoreAr
        self.hasEdgeAr = hasEdgeAr
        self.hashAr = hashAr
        self.lenOfRegionAr = lenOfRegionAr
        self.colGn1Ar = colGn1Ar
        self.colGn2Ar = colGn2Ar
        self.colEdgeAr = colEdgeAr
        self.hashArrayLen = hashArrayLen

    def returnArrays(self):
        '''Return all our arrays.'''
        return self.scoreD['rawSc'],self.hasEdgeAr,self.hashAr,self.lenOfRegionAr,self.colGn1Ar,self.colGn2Ar,self.colEdgeAr,self.hashArrayLen
        
    def endNodesToEdge(self,gn1,gn2):
        '''Given two genes, return the number of the edge between them. If
there isn't any, return None.'''

        hs = self.hashByGenePair(gn1,gn2)

        if not self.hasEdgeAr[hs]:
            # no score pairs here
            return None
        else:
            # one or more gene pairs are stored in a region
            # corresponding to this hash. check if any of them are the
            # gene pair we're after.
            stInd = self.hashAr[hs]
            regionLen = self.lenOfRegionAr[hs]
            return self.searchCollisionArrays(gn1,gn2,stInd,regionLen)

    def searchCollisionArrays(self,gn1,gn2,stInd,regionLen):
        '''Search the collision arrays for a match to gn1,gn2. If found, return index,
    otherwise return None.
        '''

        # find the first occurence of gn1, if any
        # The colGn1Ar is sorted in numerical order between stInd and stInd+regionLen
        endInd = stInd + regionLen
        searchStart = self.binarySearchGn1Array(gn1,stInd,endInd)
        if searchStart == None:
            return None

        for i in range(searchStart,endInd):
            if self.colGn1Ar[i] != gn1:
                # it's no longer gn1, thus we won't find it
                return None
            elif self.colGn2Ar[i] == gn2:
                # self.colGn1Ar[i] == gn1 since we got this far
                return self.colEdgeAr[i]
        return None
        
    def binarySearchGn1Array(self,gn1,st,end):
        '''Find the first occurrence of gn1 in self.colGn1Ar using binary
search. (colGn1Ar can be assumed to be sorted over the region st to
end).
        '''
        while st < end:
            mid = (st + end) // 2
            if self.colGn1Ar[mid] < gn1:
                st = mid + 1
            elif self.colGn1Ar[mid] > gn1:
                end = mid
            elif mid > 0 and self.colGn1Ar[mid-1] == gn1:
                end = mid
            else:
                return mid

        return None

    def getScoreByEdge(self,edge,scoreType):
        '''Given an edge get the score corresponding to scoreType.'''
        return self.scoreD[scoreType][edge]

    def getScoreByEndNodes(self,g1,g2,scoreType):
        '''Given two genes get the score corresponding to scoreType.'''

        if g1 > g2: g2,g1 = g1,g2 # make sure g1 is lower gene num
        edge = self.endNodesToEdge(g1,g2)
        if edge == None:
            return None
        else:
            return self.getScoreByEdge(edge,scoreType)



import os,sys
from Bio import SeqIO

def getUniqueRedundSets(fileName,speciesName):
    '''Run through genbank file fileName, get set of unique genes (with no
redundancies), and set with redundancies.'''
    f = open(fileName, 'rU')
    geneL=[]
    for record in SeqIO.parse(f, "genbank"):
        # iterate through the genes on the chromosome
        for feature in record.features:
            # choose only the features that are protein coding genes
            if feature.type == "CDS" and 'protein_id' in feature.qualifiers:
                geneName = speciesName + '-' + feature.qualifiers['protein_id'][0]
                geneL.append(geneName)
                
    f.close()

    # now figure out which ones are unique
    uniqueS = set()
    redundS = set()
    for gene in geneL:
        if geneL.count(gene)>1:
            redundS.add(gene)
        else:
            uniqueS.add(gene)
    return uniqueS,redundS


def parseGenbank(paramD,fastaOutFileDir,genbankFileList,fileNameMapD):
    '''We pass through each genbank file twice. Once to identify redundant
genes, so we can avoid them. And another time to get the stuff we
want.'''

    if genbankFileList == []:
        raise ValueError("List of genbank files to parse is empty.")
    
    geneOrderOutFile = open(paramD['geneOrderFN'], 'w')
    redundFile = open(paramD['redundProtsFN'], 'w')
    geneInfoFile = open(paramD['geneInfoFN'], 'w')

    problemGenbankFileL = []
    
    # iterate through list of genbank files
    for fileName in genbankFileList:

        genbankName = os.path.split(fileName)[-1]

        speciesName = fileNameMapD[genbankName]

        uniqueS,redundS=getUniqueRedundSets(fileName,speciesName)

        # Check if we've been passed a .gbff file with no protein
        # annotations, and if so tell user
        if len(uniqueS) == 0 and len(redundS) == 0:
            problemGenbankFileL.append(fileName)
            continue

        # write the redundant ones for this species to our redund file
        for gene in redundS:
            redundFile.write(gene + "\n")
            
        inFile = open(fileName, 'rU')
        fastaOutName = fastaOutFileDir + speciesName + ".fa"
        fastaOutFile = open(fastaOutName, 'w')

        # start next line in geneOrderOutFile
        geneOrderOutFile.write(speciesName)
        
        # iterate through chromosomes in the genbank file
        for record in SeqIO.parse(inFile, "genbank"):
            geneStartSeqL = []
            chrom = record.id # .id, as opposed to .name includes the version id
            #iterate through the genes on the chromosome
            for feature in record.features:
                # choose only the features that are protein coding genes
                if feature.type == "CDS" and 'protein_id' in feature.qualifiers and 'translation' in feature.qualifiers:
                    geneName = speciesName + '-' + feature.qualifiers['protein_id'][0]
                    # verify not in set of genes that appear more than once
                    if geneName in uniqueS:
                        start = int(feature.location.start)
                        end = int(feature.location.end)
                        # strand
                        if feature.location.strand == 1:
                            strand = "+"
                        elif feature.location.strand == -1:
                            strand = "-"
                        else:
                            strand = "?"
                        
                        aaSeq = feature.qualifiers['translation'][0]

                        # get common name
                        commonName=''
                        if 'gene' in feature.qualifiers:
                            commonName=feature.qualifiers['gene'][0]

                        locusTag=''
                        if 'locus_tag' in feature.qualifiers:
                            locusTag = feature.qualifiers['locus_tag'][0]
                        
                            
                        # get description
                        descrip=''
                        if 'gene' in feature.qualifiers:
                            descrip += feature.qualifiers['gene'][0]+' - '
                        if 'product' in feature.qualifiers:
                            descrip += feature.qualifiers['product'][0]

                        geneStartSeqL.append((geneName,commonName,locusTag,descrip,chrom,start,end,strand,aaSeq))

            if geneStartSeqL != []: # if its not empty
                # no need to sort, it's in order in the file
                geneL=[]
                for geneName,commonName,locusTag,descrip,chrom,start,end,strand,aaSeq in geneStartSeqL:
                    # write to fastaOutFile
                    fastaOutFile.write(">" + geneName + "\n" + aaSeq + "\n")
                    geneInfoFile.write("\t".join([geneName,commonName,locusTag,descrip,chrom,str(start),str(end),strand]) + "\n")
                    geneL.append(geneName)
                    
                # write this chromosome to geneOrderFile                    
                geneOrderOutFile.write("\t"+" ".join(geneL))
                
        geneOrderOutFile.write("\n")
                    
        inFile.close()
        fastaOutFile.close()

    geneOrderOutFile.close()
    redundFile.close()
    geneInfoFile.close()

    # If there are any files in problemGenbankFileL, throw error
    if problemGenbankFileL != []:
    
        raise ValueError("The following genbank files lack protein annotations:\n" + "\n".join(problemGenbankFileL))




import parasail,glob,statistics,sys
from multiprocessing import set_start_method, Pool
from . import genomes
from . import trees
from . import Score


#### Global variables for use with multithreading

sharedScoresO = Score.sharedScore()

#### raw similarity scores

def calcRawScores(paramD,geneNames,scoresO):
    '''Get a global alignment based raw score for every edge in scoresO.'''

    numThreads = paramD['numThreads']
    gapOpen = paramD['gapOpen']
    gapExtend = paramD['gapExtend']
    matrix = paramD['matrix']
    
    # load sequences
    protFnL=glob.glob(paramD['fastaFilePath'])
    seqD=genomes.loadProt(protFnL)

    # make list of sets of arguments to be passed to p.map. There
    # should be numThreads sets.
    argumentL = [([],seqD,gapOpen, gapExtend, matrix) for i in range(numThreads)]

    i=0
    for g1,g2 in scoresO.iterateEdgesByEndNodes():
        edgeNum = scoresO.endNodesToEdge(g1,g2)
        edgeT = edgeNum,geneNames.numToName(g1),geneNames.numToName(g2)
        argumentL[i%numThreads][0].append(edgeT)
        i+=1

    # run in multiple processes
    with Pool(processes=numThreads) as p:
        # store the results to scoresO as they come in
        for scoresL in p.imap_unordered(rawScoreGroup, argumentL):
            for edgeNum,sc in scoresL:
                scoresO.addScoreByEdge(edgeNum,sc,'rawSc')

    return scoresO

def rawScoreGroup(argT):
    '''Given a dictionary of sequences and a list of gene pairs, go
through each pair and get a needleman wunch based score.
    '''
    edgeL,seqD,gapOpen,gapExtend,matrix = argT
    matrix = eval(matrix) # turn it from string to parasail matrix object
    scoresL = []
    for edgeNum,g1,g2 in edgeL:
        scaled = rawScore(seqD[g1],seqD[g2],gapOpen,gapExtend,matrix)
        scoresL.append((edgeNum,scaled))
    return scoresL
   
def rawScore(s1,s2,gapOpen, gapExtend, matrix):
    '''Calculate score between a pair of protein sequences, based on a
global alignment. We scale the alignment score to be between 0 and 1,
based on the max and min possible scores for these sequences.'''

    # note. Here we've got a score 0-1 based only on the alignment
    # score, and not requiring the alignment itself. Right now
    # parasail doens't give the alignment. It might in the future, and
    # we could revisit this.
    
    r_s1s2 = parasail.nw_scan(s1,s2, gapOpen, gapExtend, matrix)

    if len(s1) < len(s2):
        r_self = parasail.nw_scan(s1,s1, gapOpen, gapExtend, matrix)
    else:
        r_self = parasail.nw_scan(s2,s2, gapOpen, gapExtend, matrix)

    sc = r_s1s2.score
    mx = r_self.score # max possible is shorter seq against itself.
    
    # lowest possible score, if we have gaps opposite all residues and
    # two opens. Note parasail does not count gap extend for the
    # residue where a gap is opened, hence the -2 in the extend
    # formula below.
    mn = - ( 2 * gapOpen + ( (len(s1)+len(s2) -2 ) * gapExtend ) )
    scaled = (sc - mn) / (mx - mn)
    
    return scaled


#### synteny scores

def calcSynScores(scoresO,geneNames,geneOrderT,paramD,tree):
    '''Calculate the synteny score between two genes and add to edge
attributes of scoresO. We only bother making synteny scores for those
genes that have an edge in scoresO.
    '''

    synWSize = paramD['synWSize']
    numSynToTake = paramD['numSynToTake']
    numThreads = paramD['numThreads']
    
    neighborTL = createNeighborL(geneNames,geneOrderT,synWSize)
    scoresO.initializeScoreArray('synSc') # array to store final synSc result in
    
    ## Prepare argument list
    
    # make list of groups of arguments to be passed to p.imap. There
    # should be numThreads groups.
    argumentL = [[[],tree,neighborTL,numSynToTake,geneNames] for i in range(numThreads)]

    i=0
    for gn1,gn2 in scoresO.iterateEdgesByEndNodes():
        argumentL[i%numThreads][0].append((gn1,gn2))
        i+=1

    ## prepare raw arrays to share
    sharedScoresO.createArrays(scoresO,paramD)

    rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,hashArrayLen = sharedScoresO.returnArrays()
    
    ## Run
    with Pool(processes=numThreads,initializer=synScoreGroupInit,initargs=(rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,hashArrayLen)) as p:
        for synScoresL in p.imap_unordered(synScoreGroup, argumentL):
            for gn1,gn2,sc in synScoresL:
                scoresO.addScoreByEndNodes(gn1,gn2,sc,'synSc')

    return scoresO

def createNeighborL(geneNames,geneOrderT,synWSize):
    '''Return a list which specifies the neighbors of each gene. Index of
list corresponds to gene number, and the value located at that index
is a tuple of all genes within a synWSize window. e.g. synWSize 10 means we
go 5 genes in either direction.'''

    lenInEitherDirec = int(synWSize/2)
    
    neighborTL = [None for x in geneNames.nums]
    
    for contigT in geneOrderT:
        if not contigT == None:
            for geneNumT in contigT:
                for i in range(len(geneNumT)):
                    end = i + lenInEitherDirec
                    st = i-lenInEitherDirec if i-lenInEitherDirec>0 else 0 # st can't be less than 0
                    L = list(geneNumT[st:end])
                    L.remove(geneNumT[i])
                    neighborTL[geneNumT[i]] = tuple(L)

    return neighborTL

def synScoreGroupInit(rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,hashArrayLen):
    '''Initializer for each separate process doing the synSc
calculation. Loads the global sharedScoresO object with shared
arrays.'''
    sharedScoresO.insertArrays(rawScoreAr,hasEdgeAr,hashAr,lenOfRegionAr,colGn1Ar,colGn2Ar,colEdgeAr,hashArrayLen)

    
def synScoreGroup(argsT):
    '''Given an argument list, including pairs of genes, calculate synteny
    scores. This function is intended to be called by p.map.
    '''

    edgeL,tree,neighborTL,numSynToTake,geneNames = argsT
    
    outL=[]
    for gn1,gn2 in edgeL:
        outL.append(synScore(sharedScoresO,gn1,gn2,tree,neighborTL,numSynToTake,geneNames))
        
    return outL
        
def synScore(sharedScoresO,gn1,gn2,tree,neighborTL,numSynToTake,geneNames):
    '''Given two genes, calculate a synteny score for them. We are given
    the genes, neighborTL, which contains lists of neighbors for each
    gene. For the two sets of neighbors, we find the numSynToTake top
    pairs, and return the average of their scores. The approach is
    greedy. We find the pair with the best score, add it, then remove
    those genes and iterate.
    '''

    sp1 = geneNames.numToStrainNum(gn1)
    sp2 = geneNames.numToStrainNum(gn2)

    L1 = list(neighborTL[gn1])
    L2 = list(neighborTL[gn2])

    topScL= [0] * numSynToTake # min raw score is 0

    for i in range(numSynToTake):
        ind1,ind2,sc = topScore(L1,L2,sharedScoresO)
        if sc == -float('inf'):
            break
        topScL[i] = sc
        del(L1[ind1]) # remove this index
        del(L2[ind2])
        
    synSc = sum(topScL) / numSynToTake
    
    return gn1, gn2, synSc

def topScore(L1,L2,sharedScoresO):
    '''Find the best norm score between genes in L1 and L2. Return the index of
each and the score.'''
    besti1 = 0
    besti2 = 0
    bestSc = -float('inf')

    for i1,gn1 in enumerate(L1):
        for i2,gn2 in enumerate(L2):
            sc = sharedScoresO.getScoreByEndNodes(gn1,gn2,'rawSc')
            if sc != None and sc > bestSc:
                bestSc = sc
                besti1 = i1
                besti2 = i2
    return besti1,besti2,bestSc


#### Calculating the set of all around best reciprocal hit homologs
#    (used in core synteny scores below)

def createAabrhL(blastFilePath,strainNamesL,evalueThresh,aabrhFN):
    '''Get the sets of all around best reciprocal hits.'''

    blastDir = blastFilePath.split("*")[0]
    rHitsL=getAllReciprocalHits(blastDir,strainNamesL,evalueThresh)
    
    # get sets of genes in first species that have a reciprocal best
    # hit in each other species. this is to save time in next step.
    possibleGenesD = getPossibleGenesD(rHitsL)

    # then, out of the possible genes, get those that are all
    # pairwise reciprocal bets hits
    aabrhL = getAllAroundBRHL(possibleGenesD, rHitsL)

    # write it to file
    f=open(aabrhFN,'w')
    for orthoT in aabrhL:
        f.write("\t".join(orthoT)+"\n")
    f.close()
        
    return aabrhL


def loadOrthos(aabrhFN):
    '''Reads the all around best reciprocal hits orthologs file. One set
per line.'''
    f= open(aabrhFN,'r')

    orthoL=[]
    while True:
        s=f.readline()
        if s=='':
            break
        L=s.rstrip().split('\t')
        orthoL.append(tuple(L))
            
    f.close()
    return orthoL

def getAllReciprocalHits(blastDir,strainNamesL,evalueThresh):
    '''return an upper-diagonal (N-1)xN matrix where each entry
    [i][j] (j > i) contains a dictionary of best reciprocal hits
    between species i and species j, as indexed in list strainNamesL;
    key is always gene in species i, and value gene in species j'''

    rHitsL = []
    
    for i in range(len(strainNamesL)): 
        rHitsL.append([])
        # fill in the rest of the row with the dictionary of best
        # reciprocal hits between species i and j (keyed by species i)
        for j in range(len(strainNamesL)):
            if j > i:
                rHitsL[i].append(getReciprocalHits(strainNamesL[i], strainNamesL[j], blastDir, evalueThresh))
            else:
                rHitsL[i].append(None)
    return rHitsL


def getReciprocalHits(strainName1, strainName2, blastDir, evalueThresh):
    '''Given strain names and blast file directory name, load hits between
two strains in each direction, then go through and keep only the
reciprocal hits. Returns dictionary where keys are genes in strain 1
and values are corresponding genes in strain 2.'''
    # get best hits of 1 blasted against 2...
    hits1D = getHits(blastDir+strainName1+'_-VS-_'+strainName2+'.out', evalueThresh)
    # ...and 2 blasted against 1
    hits2D = getHits(blastDir+strainName2+'_-VS-_'+strainName1+'.out', evalueThresh)

    # then store only the reciprocal best hits
    recipHitsD = {}
    for query in hits1D:
        hit = hits1D[query]
        nextHit = hits2D.get(hit)
        if nextHit == query:
            recipHitsD[query] = hit

    return recipHitsD

def getHits(fileName, evalueThresh):
    """Given a BLAST output file, returns a dictionary keyed by the genes
    in the query species, with the values being the top hit (if any)
    for those genes. Assumes the blast hits for each query are given
    from most to least significant. which appears to be the
    case. Thresholds for minimum similarity and maximum length
    difference are globally defined.
    """

    f = open(fileName, 'r')
    hitsD = {}

    queryGene = ""

    # read through entire file
    line = f.readline()

    while line != '':
        # break the line up into its components
        L = line.split('\t')

        if len(L)<12:
            # its not an info line (likely header)
            line = f.readline()
            continue

        # gene names come first
        queryGene = L[0]
        hit = L[1]
        evalue = float(L[10])

        if evalue < evalueThresh:
            # store the query and hit if they meet thresholds
            hitsD[queryGene] = hit

        # we only want the first hit for any query gene. Change this
        # later to check to get best score even if doesn't come first.
        while line.split('\t')[0] == queryGene:
            line = f.readline()

    f.close()
    return hitsD

def getPossibleGenesD(rHitsL):
    '''return dict containing an entry for each gene in species 0 that has
    a best reciprocal hit in all other species.  format of an entry is
    gene:(tuple of recip hit genes in other species). We do this to
    quickly eliminate things we aren't interested in.
    '''
    possibleGenesD = {}
    firstRow = rHitsL[0]
    firstDict = firstRow[1]
    # we just look at first dict, since the all around brh sets will
    # have a gene in this with matches in all strains.
    for gene in firstDict:
        isPossible = True
        hitsL = [firstDict[gene]]
        for colInd in range(2, len(firstRow)):
            compareDict = firstRow[colInd]
            hit = compareDict.get(gene)
            if hit == None:
                isPossible = False
                break
            else:
                hitsL.append(hit)
        if isPossible:
            possibleGenesD[gene] = tuple(hitsL)
    return possibleGenesD


def getAllAroundBRHL(possibleGenesD, rHitsL):
    '''Given a matrix of reciprocal hit dictionaries in rHitsL, and a dict
of possible hits, return a list of all around best reciprocal hits. In
the form of a list of lists, where each sublist is one set of
orthologs.'''
    
    aabrhL = []

    genesL = list(possibleGenesD.keys())
    genesL.sort() # sort so we get consistent order
    
    for gene in genesL:
        hitsT = possibleGenesD[gene]

        # check that each gene in the list of hitsT is best reciprocal 
        # hit of all following genes
        isOrth = True
        for hit1Index in range(len(hitsT)):
            hit1 = hitsT[hit1Index]

            # check that current gene in list is best reciprocal hit
            # of all following genes in list
            for hit2Index in range(hit1Index+1, len(hitsT)): # j always > i
                hit2 = hitsT[hit2Index]
                sp1sp2Dict = rHitsL[hit1Index+1][hit2Index+1]
                match = sp1sp2Dict.get(hit1)
                if match != hit2: # oh noes! not reciprocal!
                    isOrth = False
                    break

            # if any pair not reciprocal best hits, then we fail
            if not isOrth:
                break

        # if we get through all genes without finding that some aren't
        # reciprocal best hits, then we keep this list of genes
        if isOrth:
            aabrhL.append((gene,)+hitsT)
    return aabrhL


#### Core synteny scores

def calcCoreSynScores(scoresO,strainNamesL,paramD,geneNames,geneOrderT):
    '''Calculate synteny scores based on core genes given in
aabrhL. Scores are between 0 and 1, giving the percentage of syntenic
genes shared.'''

    blastFilePath = paramD['blastFilePath']
    evalueThresh = paramD['evalueThresh']
    aabrhFN = paramD['aabrhFN']
    coreSynWsize = paramD['coreSynWsize']
    
    aabrhL = createAabrhL(blastFilePath,strainNamesL,evalueThresh,aabrhFN)
    
    geneToAabrhT = createGeneToAabrhT(aabrhL,geneNames)
    coreSyntenyT = createCoreSyntenyT(geneToAabrhT,geneOrderT,coreSynWsize)

    # Not parallelized. Pretty fast already.

    # loop over all edges in scoresO, adding corresponding core syn score
    scoresO.initializeScoreArray('coreSynSc') # create array
    for gn1,gn2 in scoresO.iterateEdgesByEndNodes():
        coreSynSc=coreSynScore(coreSyntenyT[gn1],coreSyntenyT[gn2],coreSynWsize)
        scoresO.addScoreByEndNodes(gn1,gn2,coreSynSc,'coreSynSc')
        
    return scoresO

def createGeneToAabrhT(aabrhL,geneNames):
    '''Create a tuple where the index corresponds to gene number and the
value at that location is the number of the aabrh group to which the
gene belongs, or None. The aabrh groups are numbered, simply based on
the index where they occur in aabrhL.'''

    geneToAabrhL =[None] * len(geneNames.nums)

    for aabrhNum in range(len(aabrhL)):
        for geneName in aabrhL[aabrhNum]:
            geneNum = geneNames.nameToNum(geneName)
            geneToAabrhL[geneNum] = aabrhNum
    return tuple(geneToAabrhL)

def createCoreSyntenyT(geneToAabrhT,geneOrderT,coreSynWsize):
    '''Create and return core synteny tuple. The index of this corresponds
to gene number. The value at that index is a tuple of the aabrh
numbers for the syntenic genes within coreSynWsize.'''

    coreSyntenyL =[None] * len(geneToAabrhT)
    geneAabrhOrderL = createGeneAabrhOrderL(geneToAabrhT,geneOrderT)
    for element in geneAabrhOrderL:
        if element != None:
            # internal nodes are None, having no genes themselves
            for geneNumT,aabrhNumT in element:
                for pos in range(len(geneNumT)):
                    geneNum=geneNumT[pos]
                    coreSyntenyL[geneNum]=getAabrhContext(pos,aabrhNumT,coreSynWsize)
    return tuple(coreSyntenyL)
                    
def getAabrhContext(pos,aabrhNumT,coreSynWsize):
    '''Starting at index pos in aabrhNumT, return the coreSynWsize/2 aabrh
group numbers to each side (not including the gene at pos).'''

    numToTakeEachSide = int(coreSynWsize/2)
    coreSynL=[]

    # go forward from gene
    ct=0
    j=pos+1
    while j < len(aabrhNumT) and ct < numToTakeEachSide:
        if aabrhNumT[j]!=None:
            coreSynL.append(aabrhNumT[j])
            ct+=1
        j+=1

    # go backward from gene
    ct=0
    j=pos-1
    while j >= 0 and ct < numToTakeEachSide:
        if aabrhNumT[j]!=None:
            coreSynL.append(aabrhNumT[j])
            ct+=1
        j-=1

    return tuple(coreSynL)
            
def createGeneAabrhOrderL(geneToAabrhT,geneOrderT):
    '''Make geneAabrhOrderL, which follows the structure of geneOrderT,
but instead of only having geneNumT's, it has (geneNumT,aabrhNumT)
pairs at its lowest level.
    '''
    geneAabrhOrderL=[]
    for contigT in geneOrderT:
        if contigT == None:
            geneAabrhOrderL.append(None)
        else:
            contigL=[]
            for geneNumT in contigT:
                aabrhNumT = convertGeneToAabrhTuple(geneNumT,geneToAabrhT)
                contigL.append((geneNumT,aabrhNumT))
            geneAabrhOrderL.append(tuple(contigL))
    return geneAabrhOrderL

def convertGeneToAabrhTuple(geneNumT,geneToAabrhT):
    '''Take a tuple of gene numbers and make a corresponding tuple where
every gene number is replaced by its corresponding aabrh number, or
None if it has none.'''
    aabrhNumL=[]
    for geneNum in geneNumT:
        aabrhNumL.append(geneToAabrhT[geneNum])
    return tuple(aabrhNumL)

def coreSynScore(synT1,synT2,coreSynWsize):
    '''Given two tuples, determine the number of shared core genes in
each. Return this, divided by the core synteny window size, which is
the number of genes we consider around each gene. Values will be
between 0 and 1.'''
    ct=0
    for cgene in synT1:
        if cgene in synT2:
            ct+=1
    return ct/coreSynWsize


#### Score I/O

def writeScores(scoresO,geneNames,scoresFN):
    '''Write a scores object to file. If scoresFN has the .bout extension, write
binary version, otherwise write in text output format.'''

    scoreTypeL = ['rawSc','synSc','coreSynSc']
    if scoresFN.split('.')[-1] == 'bout':
        scoresO.writeScoresBinary(scoreTypeL,scoresFN)
    else:
        scoresO.writeScoresText(geneNames,scoreTypeL,scoresFN)


def readScores(scoresFN,geneNames=None):
    '''Read scores from file creating a Score object of scores. If
scoresFN has the .bout extension, read binary, otherwise read text
format.
    '''

    scoreTypeL = ['rawSc','synSc','coreSynSc']
    if scoresFN.split('.')[-1] == 'bout':
        scoresO = Score.Score.readScoresBinary(scoreTypeL,scoresFN)
    else:
        scoresO = Score.Score.readScoresText(geneNames,scoreTypeL,scoresFN)
    return scoresO




class LocusFamily:
    def __init__(self, famNum, locusFamNum, lfMrca):
        '''Initialize a LocusFamily object with a family number, a LocusFamily
number, and the mrca for this locus family (which may differ from the
mrca for its family.'''
        self.famNum = famNum
        self.locusFamNum = locusFamNum
        self.lfMrca = lfMrca
        self.geneD = {}

    def addGene(self, gene,geneNames):
        '''Add a gene (in numerical form) to this LocusFamily.'''
        strain = geneNames.numToStrainNum(gene)
        if strain in self.geneD:
            self.geneD[strain].append(gene)
        else:
            self.geneD[strain] = [gene]
        
    def addGenes(self,genesL,geneNames):
        '''Add a list (or set etc.) of genes to this LocusFamily.'''
        for gene in genesL:
            self.addGene(gene,geneNames)

    def iterGenes(self):
        '''Iterate through all genes in this locus family.'''
        for L in self.geneD.values():
            for gene in L:
                yield gene

    def iterGenesByStrain(self,strain):
        '''Iterate through all genes in this locus family that are in a
        particular strain.
        '''
        if strain in self.geneD:
            for gene in self.geneD[strain]:
                yield gene
        return

    def iterStrains(self):
        '''Iterator that yields all the strains in which this locus family
occurs.'''
        for strain in self.geneD.keys():
            yield strain
    
    def getStr(self,strainNum2StrD,geneNames,sep):
        '''Return a string representation of a single LocusFamily. Separator
between elements given by sep. Elements are: locusFamNum lfMrca gene1
gene2...
        '''
        outL=[str(self.locusFamNum),strainNum2StrD[self.lfMrca]]
        
        for geneNum in self.iterGenes():
            outL.append(geneNames.numToName(geneNum))

        return sep.join(outL)
    
    def fileStr(self,strainNum2StrD,geneNames):
        '''Return a string representation of a single LocusFamily. Format is
        comma separated: 
        '''
        return self.getStr(strainNum2StrD,geneNames,',')

    def __repr__(self):
        '''String representation of a LocusFamily, for display purposes.'''
        return "<lf "+str(self.locusFamNum)+">"
        
        
class Family:
    def __init__(self,famNum,mrca,seedPairL=None):
        '''Initialize an object of class Family.'''

        self.famNum = famNum
        self.mrca = mrca
        self.seedPairL = seedPairL # original seeds from PHiGS. This will be
                                   # None if single gene family
        self.locusFamiliesL = []   # will contain locusFamily objects for this family

    def addLocusFamily(self,lfO):
        self.locusFamiliesL.append(lfO)

    def getLocusFamilies(self):
        return self.locusFamiliesL

    def iterGenes(self):
        '''Iterate through all the genes associated with this family in numerical form.'''
        for lfO in self.getLocusFamilies():
            for gene in lfO.iterGenes():
                yield gene

    def iterStrains(self):
        '''Iterator that yields all the strains in which this family
occurs.
        '''
        for lfO in self.getLocusFamilies():
            for strain in lfO.iterStrains():
                yield strain
                
    def getOutsideConnections(self,scoresO):
        '''Given a score object, return a set of all outside genes with
connections to this family.'''
        allGenesInFamS = set(self.iterGenes())
        otherGenesS=set()
        for geneNum in allGenesInFamS:
            for otherGene in scoresO.getConnectionsGene(geneNum):
                if not otherGene in allGenesInFamS:
                    otherGenesS.add(otherGene)
        return otherGenesS
    
    def fileStr(self,strainNum2StrD,geneNames):
        '''Return string representation of single family. Format is: famNum <tab> 
        mrca <tab> seedG1 <tab> seedG2 <tab> locusFamNum1,locusFamGenes <tab>
        locusFamNum2,locusFamGenes...
        The LocusFamily object representations are comma separated.
        '''
        
        outL =[str(self.famNum)]
        outL.append(strainNum2StrD[self.mrca])

        if self.seedPairL == None:
            outL.extend(["-","-"])
        else:
            for seed in self.seedPairL:
                outL.append(geneNames.numToName(seed))

        for lfO in self.locusFamiliesL:
            outL.append(lfO.fileStr(strainNum2StrD,geneNames))
                
        return "\t".join(outL)

    def __repr__(self):
        return "<fam:"+str(self.famNum)+">"
    

class Families:

    def __init__(self,tree):
        '''Initialize an object of class Families.'''

        self.tree = tree
        self.locusFamiliesD = {}
        self.familiesD = {} 

        ## locusFamiliesD has key locusFamNum and value a LocusFamily
        ## object. familiesD has key famNum and value
        ## [mrca,seedG1,seedG2,[List of lf numbers]]

        
    def initializeFamily(self,famNum,mrca,seedPairL=None):
        '''Set up an entry for family famNum.'''
        # the seed genes are the original PHiGs seed.

        self.familiesD[famNum] = Family(famNum,mrca,seedPairL=seedPairL)

    def addLocusFamily(self, lfO):
        '''Add a LocusFamily. Assumes initializeFamily has already been called
to create the corresponding family.
        '''
        self.locusFamiliesD[lfO.locusFamNum] =  lfO
        self.familiesD[lfO.famNum].addLocusFamily(lfO)

    def getLocusFamily(self,locusFamNum):
        return self.locusFamiliesD[locusFamNum]

    def getFamily(self,famNum):
        return self.familiesD[famNum]

    def iterLocusFamilies(self):
        '''Iterate over all LocusFamilies in order of locusFamNum.
        '''
        maxLocusFamNum = max(self.locusFamiliesD.keys())
        for i in range(maxLocusFamNum+1):
            if i in self.locusFamiliesD:
                yield self.locusFamiliesD[i]
        
    def iterFamilies(self):
        '''Iterate over all Families in order of famNum.'''
        maxFamNum = max(self.familiesD.keys())
        for i in range(maxFamNum+1):
            if i in self.familiesD:
                yield self.familiesD[i]
                
    def getAllGenes(self):
        '''Collect all the genes present in the LocusFamily objects belonging
to this instance of the Families class. Return as a set.
        '''
        allGenesS=set()
        for lfO in self.iterLocusFamilies():
            allGenesS.update(set(lfO.iterGenes()))
        return allGenesS

    def __repr__(self):
        return "<Families object--"+str(len(self.familiesD))+" Families, "+str(len(self.locusFamiliesD))+" LocusFamilies>"
